import torch
import torch.nn as nn
import torch.nn.functional as F

from e2eAIOK.DeNas.module.layernorm_base import LayerNormBase


class LayerNormSuper(LayerNormBase):
    def __init__(self, super_embed_dim):
        super().__init__(super_embed_dim)

        # the largest embed dim
        self.super_embed_dim = super_embed_dim

        # the current sampled embed dim
        self.sample_embed_dim = None

        self.samples = {}
        self.profiling = False

    def profile(self, mode=True):
        self.profiling = mode

    def sample_parameters(self, resample=False):
        if self.profiling or resample:
            return self._sample_parameters()
        return self.samples

    def _sample_parameters(self):
        self.samples['weight'] = self.weight[:self.sample_embed_dim]
        self.samples['bias'] = self.bias[:self.sample_embed_dim]
        return self.samples

    def set_sample_config(self, sample_embed_dim):
        self.sample_embed_dim = sample_embed_dim
        self._sample_parameters()

    def forward(self, x):
        self.sample_parameters()
        return F.layer_norm(x, (self.sample_embed_dim,), weight=self.samples['weight'], bias=self.samples['bias'], eps=self.eps)

    def calc_sampled_param_num(self):
        assert 'weight' in self.samples.keys()
        assert 'bias' in self.samples.keys()
        return self.samples['weight'].numel() + self.samples['bias'].numel()

    def get_complexity(self, sequence_length):
        return sequence_length * self.sample_embed_dim



import os
import torch
from torchvision import datasets, transforms
from e2eAIOK.common.trainer.data.data_builder_cv import DataBuilderCV
from e2eAIOK.common.utils import check_mkdir

class DataBuilderCIFAR(DataBuilderCV):
    def __init__(self, cfg):
        super().__init__(cfg)
    
    def prepare_dataset(self):
        """
            prepare CIFAR dataset
        """
        if self.cfg.data_set.lower() in ["cifar10","cifar100"]:
            dataset_train = self.build_dataset(is_train=True)
            dataset_val = self.build_dataset(is_train=False)
        else:
            raise RuntimeError(f"dataset {self.cfg.data_set} not supported")
        self.dataset_train = dataset_train
        self.dataset_val = dataset_val
    
    def build_dataset(self, is_train = True):
        transform = self.build_transform(is_train)
        data_folder = check_mkdir(self.cfg.data_path)
        if self.cfg.data_set.lower() == 'cifar10':
            dataset = datasets.CIFAR10(data_folder, train=is_train, transform=transform, download=True)
        elif self.cfg.data_set.lower() == 'cifar100':
            dataset = datasets.CIFAR100(data_folder, train=is_train, transform=transform, download=True)
        return dataset
    def build_transform(self, is_train):
        resize_im = "input_size" in self.cfg and self.cfg.input_size > 32
        if is_train:
            # this should always dispatch to transforms_imagenet_train
            transform = transforms.Compose([
                transforms.RandomCrop(32, padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
            ])
            if not resize_im:
                # replace RandomResizedCropAndInterpolation with
                # RandomCrop
                transform.transforms[0] = transforms.RandomCrop(
                    self.cfg.input_size, padding=4)
            return transform
        else:
            transform = transforms.Compose([
                transforms.ToTensor(),
                transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
            ])

        return transform


#!/usr/bin/python
# -*- coding: utf-8 -*-
# @Author : Wang Xinyao      
# @Time   : 10/20/2022

import torch
import torch.nn
from .dataset_wrapper import DatasetWrapper
import datetime, time
import os
import numpy as np
import logging

def logits_wrap_dataset(dataset, logits_path, num_classes, save_logits, topk=0):
    dataset = DatasetWrapper(dataset,
                                logits_path=logits_path,
                                num_classes = num_classes,
                                topk=topk,
                                write=save_logits)
    return dataset

def save_check_logits(model, dataloader, epochs, start_epoch=0, topk=0, num_classes=10, model_type=None, device="cpu", save_flag=True, check_flag=False):
    if save_flag and check_flag:
        raise RuntimeError("Can not save and check logits together!")

    epoch_steps = len(dataloader)
    dataset = dataloader.dataset
    for epoch in range(start_epoch, epochs, 1):
        start_time = time.time()
        dataset.set_epoch(epoch)
        if save_flag:
            save_logits_epoch(model, dataloader, epoch_steps, topk,  model_type, device)
        if check_flag:
            check_logits_epoch(model, dataloader, epoch_steps, topk, num_classes, model_type, device)
        print(f"Epoch {epoch} took {time.time()-start_time} seconds")
        logging.info(f"Epoch {epoch} took {time.time()-start_time} seconds")

def save_logits_epoch(model, dataloader, epoch_steps, topk=0, model_type=None, device="cpu"):
    ''' save teacher logits for distiller

    :param model: the pretrained teacher model
    :param dataloader: dataloader
    :param epoch_steps: steps of one epoch
    :param topk: save top k logits, 0 means save all logits
    :param model_type: type of model
    :param device: running on cpu or gpu
    '''
    with torch.no_grad():
        model.eval()
        logits_manager = dataloader.dataset.get_manager()

        #################### iterate on dataset ##############
        for idx, ((data, label), (keys, seeds)) in enumerate(dataloader):
            data = data.to(device)
            label = label.to(device)
            output = model(data)
            output = output[0] if isinstance(output, tuple) else output
            output = output.logits if model_type is not None and model_type.startswith("huggingface") else output
            
            if topk == 0: # save all logits
                values = output.detach().to(device='cpu', dtype=torch.float32)

                seeds = seeds.numpy()
                values = values.numpy()
                assert seeds.dtype == np.int32, seeds.dtype
                assert values.dtype == np.float32, values.dtype

                for key, seed, value in zip(keys, seeds, values):
                    bstr = seed.tobytes() + value.tobytes()
                    logits_manager.write(key, bstr)
            elif topk > 0: # only save topk logits
                softmax_prob = torch.softmax(output, -1)
                values, indices = softmax_prob.topk(k=topk, dim=-1, largest=True, sorted=True)
                values = values.detach().to(device='cpu', dtype=torch.float16)
                indices = indices.detach().to(device='cpu', dtype=torch.int16)

                seeds = seeds.numpy()
                values = values.numpy()
                indices = indices.numpy()
                assert seeds.dtype == np.int32, seeds.dtype
                assert indices.dtype == np.int16, indices.dtype
                assert values.dtype == np.float16, values.dtype

                for key, seed, indice, value in zip(keys, seeds, indices, values):
                    bstr = seed.tobytes() + value.tobytes() + indice.tobytes()
                    logits_manager.write(key, bstr)
            if idx % 10 == 0:
                dt = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S') # date time
                print(f"{dt} save {idx}/{epoch_steps}")
                logging.info(f"{dt} save {idx}/{epoch_steps}")

def load_logits(save_values,topk=0, num_classes=10):
    if topk == 0:
        logits_value, _ = save_values
        logits_value = logits_value.float()
    elif topk > 0:
        logits_value, logits_index, _ = save_values
        logits_index = logits_index.long()
        logits_value = logits_value.float()
        minor_value = (1.0 - logits_value.sum(-1, keepdim=True)
                    ) / (num_classes - topk)
        minor_value = minor_value.repeat_interleave(num_classes, dim=-1)
        logits_value = minor_value.scatter_(-1, logits_index, logits_value)
    return logits_value

def check_logits_epoch(model, dataloader, topk=0, num_classes=10, model_type=None, device="cpu"):
    ''' check saved teacher logits for distiller

    :param model: the evaluated model
    :param dataloader: dataloader
    :param topk: top k logits saved, 0 means save all logits
    :param num_classes: number of prediction classes
    :param model_type: type of model
    :param device: running on cpu or gpu
    '''
    with torch.no_grad():
        model.eval()  # set evaluating flag
        metric_values = {}
        #################### iterate on dataset ##############
        for idx, (data, label) in enumerate(dataloader):
            inputs, save_values = data
            inputs = inputs.to(device)
            label = label.to(device)
            batch_size = inputs.size(0)
            output = model(inputs)
            output = output[0] if isinstance(output, tuple) else output
            output = output.logits if model_type.startswith("huggingface") else output

            logits_value = output.detach().to(device='cpu', dtype=torch.float32)
            save_logits_value = load_logits(save_values, topk, num_classes)
            # softmax_prob = torch.softmax(output, -1)
            # logits_value_topk, logits_indices_topk = softmax_prob.topk(k=topk, dim=-1, largest=True, sorted=True)
            # logits_value_topk = logits_value_topk.detach().to(device='cpu', dtype=torch.float16)
            # logits_indices_topk = logits_indices_topk.detach().to(device='cpu', dtype=torch.int16)
            # metric_values["indices_diff"] = torch.count_nonzero((logits_indices_topk != save_logits_index_topk)).item() / batch_size
            # metric_values["value_topk_diff"] = (logits_value_topk - save_logits_value_topk).abs().sum().item() / batch_size
            metric_values["value_diff"] = (logits_value - save_logits_value).abs().sum().item() / batch_size
            for key in metric_values:
                print(f"{key}: {metric_values}")


'''
Copyright (C) 2010-2021 Alibaba Group Holding Limited.
'''

import os, sys
sys.path.append(os.path.dirname(os.path.abspath(__file__)))


import numpy as np
import torch, argparse
from torch import nn
import torch.nn.functional as F
import PlainNet
from PlainNet import global_utils
from PlainNet import basic_blocks, super_blocks, SuperResKXKX, SuperResK1KXK1, SuperResIDWEXKX
import itertools

class DeSearchSpaceXXBL(object):
    seach_space_block_type_list_list = [
        [SuperResK1KXK1.SuperResK1K3K1, SuperResK1KXK1.SuperResK1K5K1, SuperResK1KXK1.SuperResK1K7K1],
        [SuperResKXKX.SuperResK3K3, SuperResKXKX.SuperResK5K5, SuperResKXKX.SuperResK7K7],
    ]

    __block_type_round_channels_base_dict__ = {
            SuperResKXKX.SuperResK3K3: 8,
            SuperResKXKX.SuperResK5K5: 8,
            SuperResKXKX.SuperResK7K7: 8,
            SuperResK1KXK1.SuperResK1K3K1: 8, SuperResK1KXK1.SuperResK1K5K1: 8, SuperResK1KXK1.SuperResK1K7K1: 8,
        }

    __block_type_min_channels_base_dict__ = {
            SuperResKXKX.SuperResK3K3: 8,
            SuperResKXKX.SuperResK5K5: 8,
            SuperResKXKX.SuperResK7K7: 8,
            SuperResK1KXK1.SuperResK1K3K1: 8,
            SuperResK1KXK1.SuperResK1K5K1: 8,
            SuperResK1KXK1.SuperResK1K7K1: 8,
        }

    @classmethod
    def _get_select_student_channels_list(cls, out_channels):
        the_list = [out_channels * 2.5, out_channels * 2, out_channels * 1.5, out_channels * 1.25,
                    out_channels,
                    out_channels / 1.25, out_channels / 1.5, out_channels / 2, out_channels / 2.5]
        the_list = [min(2048, max(8, x)) for x in the_list]
        the_list = [global_utils.smart_round(x, base=8) for x in the_list]
        the_list = list(set(the_list))
        the_list.sort(reverse=True)
        return the_list

    @classmethod
    def _get_select_student_sublayers_list(cls, sub_layers):
        the_list = [sub_layers,
                    sub_layers + 1, sub_layers + 2,
                    sub_layers - 1, sub_layers - 2, ]
        the_list = [max(0, round(x)) for x in the_list]
        the_list = list(set(the_list))
        the_list.sort(reverse=True)
        return the_list

    @classmethod
    def gen_search_space(cls, block_list, block_id):
        the_block = block_list[block_id]
        student_blocks_list_list = []

        if isinstance(the_block, super_blocks.SuperConvKXBNRELU):
            student_blocks_list = []
            student_out_channels_list = cls._get_select_student_channels_list(the_block.out_channels)
            for student_out_channels in student_out_channels_list:
                tmp_block_str = type(the_block).__name__ + '({},{},{},1)'.format(
                    the_block.in_channels, student_out_channels, the_block.stride)
                student_blocks_list.append(tmp_block_str)
            pass
            student_blocks_list = list(set(student_blocks_list))
            assert len(student_blocks_list) >= 1
            student_blocks_list_list.append(student_blocks_list)
        else:
            for student_block_type_list in cls.seach_space_block_type_list_list:
                student_blocks_list = []
                student_out_channels_list = cls._get_select_student_channels_list(the_block.out_channels)
                student_sublayers_list = cls._get_select_student_sublayers_list(sub_layers=the_block.sub_layers)
                student_bottleneck_channels_list = cls._get_select_student_channels_list(the_block.bottleneck_channels)
                for student_block_type in student_block_type_list:
                    for student_out_channels, student_sublayers, student_bottleneck_channels in itertools.product(
                            student_out_channels_list, student_sublayers_list, student_bottleneck_channels_list):

                        # filter smallest possible channel for this block type
                        min_possible_channels = cls.__block_type_round_channels_base_dict__[student_block_type]
                        channel_round_base = cls.__block_type_round_channels_base_dict__[student_block_type]
                        student_out_channels = global_utils.smart_round(student_out_channels, channel_round_base)
                        student_bottleneck_channels = global_utils.smart_round(student_bottleneck_channels,
                                                                            channel_round_base)

                        if student_out_channels < min_possible_channels or student_bottleneck_channels < min_possible_channels:
                            continue
                        if student_sublayers <= 0:  # no empty layer
                            continue
                        tmp_block_str = student_block_type.__name__ + '({},{},{},{},{})'.format(
                            the_block.in_channels, student_out_channels, the_block.stride, student_bottleneck_channels,
                            student_sublayers)
                        student_blocks_list.append(tmp_block_str)
                    pass
                    student_blocks_list = list(dict.fromkeys(student_blocks_list))
                    assert len(student_blocks_list) >= 1
                    student_blocks_list_list.append(student_blocks_list)
                pass
            pass  # end for student_block_type_list in seach_space_block_type_list_list:
        pass
        return student_blocks_list_list

class DeSearchSpaceIDWEXKX:
    def __init__(self):
        self.seach_space_block_type_list_list = [
            [SuperResIDWEXKX.SuperResIDWE1K3, SuperResIDWEXKX.SuperResIDWE2K3, SuperResIDWEXKX.SuperResIDWE4K3,
            SuperResIDWEXKX.SuperResIDWE6K3,
            SuperResIDWEXKX.SuperResIDWE1K5, SuperResIDWEXKX.SuperResIDWE2K5, SuperResIDWEXKX.SuperResIDWE4K5,
            SuperResIDWEXKX.SuperResIDWE6K5,
            SuperResIDWEXKX.SuperResIDWE1K7, SuperResIDWEXKX.SuperResIDWE2K7, SuperResIDWEXKX.SuperResIDWE4K7,
            SuperResIDWEXKX.SuperResIDWE6K7],
        ]

        self.__block_type_round_channels_base_dict__ = {
            SuperResIDWEXKX.SuperResIDWE1K3: 8,
            SuperResIDWEXKX.SuperResIDWE2K3: 8,
            SuperResIDWEXKX.SuperResIDWE4K3: 8,
            SuperResIDWEXKX.SuperResIDWE6K3: 8,
            SuperResIDWEXKX.SuperResIDWE1K5: 8,
            SuperResIDWEXKX.SuperResIDWE2K5: 8,
            SuperResIDWEXKX.SuperResIDWE4K5: 8,
            SuperResIDWEXKX.SuperResIDWE6K5: 8,
            SuperResIDWEXKX.SuperResIDWE1K7: 8,
            SuperResIDWEXKX.SuperResIDWE2K7: 8,
            SuperResIDWEXKX.SuperResIDWE4K7: 8,
            SuperResIDWEXKX.SuperResIDWE6K7: 8,
        }

        self.__block_type_min_channels_base_dict__ = {
            SuperResIDWEXKX.SuperResIDWE1K3: 8,
            SuperResIDWEXKX.SuperResIDWE2K3: 8,
            SuperResIDWEXKX.SuperResIDWE4K3: 8,
            SuperResIDWEXKX.SuperResIDWE6K3: 8,
            SuperResIDWEXKX.SuperResIDWE1K5: 8,
            SuperResIDWEXKX.SuperResIDWE2K5: 8,
            SuperResIDWEXKX.SuperResIDWE4K5: 8,
            SuperResIDWEXKX.SuperResIDWE6K5: 8,
            SuperResIDWEXKX.SuperResIDWE1K7: 8,
            SuperResIDWEXKX.SuperResIDWE2K7: 8,
            SuperResIDWEXKX.SuperResIDWE4K7: 8,
            SuperResIDWEXKX.SuperResIDWE6K7: 8,
        }

    def _get_select_student_channels_list(self, out_channels):
        the_list = [out_channels * 2.5, out_channels * 2, out_channels * 1.5, out_channels * 1.25,
                    out_channels,
                    out_channels / 1.25, out_channels / 1.5, out_channels / 2, out_channels / 2.5]
        the_list = [max(8, x) for x in the_list]
        the_list = [global_utils.smart_round(x, base=8) for x in the_list]
        the_list = list(set(the_list))
        the_list.sort(reverse=True)
        return the_list

    def _get_select_student_sublayers_list(self, sub_layers):
        the_list = [sub_layers,
                    sub_layers + 1, sub_layers + 2,
                    sub_layers - 1, sub_layers - 2, ]
        the_list = [max(0, round(x)) for x in the_list]
        the_list = list(set(the_list))
        the_list.sort(reverse=True)
        return the_list

    def gen_search_space(self, block_list, block_id):
        the_block = block_list[block_id]
        student_blocks_list_list = []

        if isinstance(the_block, super_blocks.SuperConvKXBNRELU):
            student_blocks_list = []

            if the_block.kernel_size == 1:  # last fc layer, never change fc
                student_out_channels_list = [the_block.out_channels]
            else:
                student_out_channels_list = self.get_select_student_channels_list(the_block.out_channels)

            for student_out_channels in student_out_channels_list:
                tmp_block_str = type(the_block).__name__ + '({},{},{},1)'.format(
                    the_block.in_channels, student_out_channels, the_block.stride)
                student_blocks_list.append(tmp_block_str)
            pass
            student_blocks_list = list(set(student_blocks_list))
            assert len(student_blocks_list) >= 1
            student_blocks_list_list.append(student_blocks_list)
        else:
            for student_block_type_list in self.seach_space_block_type_list_list:
                student_blocks_list = []
                student_out_channels_list = self._get_select_student_channels_list(the_block.out_channels)
                student_sublayers_list = self._get_select_student_sublayers_list(sub_layers=the_block.sub_layers)
                student_bottleneck_channels_list = self._get_select_student_channels_list(the_block.bottleneck_channels)
                for student_block_type in student_block_type_list:
                    for student_out_channels, student_sublayers, student_bottleneck_channels in itertools.product(
                            student_out_channels_list, student_sublayers_list, student_bottleneck_channels_list):

                        # filter smallest possible channel for this block type
                        min_possible_channels = self.__block_type_round_channels_base_dict__[student_block_type]
                        channel_round_base = self.__block_type_round_channels_base_dict__[student_block_type]
                        student_out_channels = global_utils.smart_round(student_out_channels, channel_round_base)
                        student_bottleneck_channels = global_utils.smart_round(student_bottleneck_channels,
                                                                            channel_round_base)

                        if student_out_channels < min_possible_channels or student_bottleneck_channels < min_possible_channels:
                            continue
                        if student_sublayers <= 0:  # no empty layer
                            continue
                        tmp_block_str = student_block_type.__name__ + '({},{},{},{},{})'.format(
                            the_block.in_channels, student_out_channels, the_block.stride, student_bottleneck_channels,
                            student_sublayers)
                        student_blocks_list.append(tmp_block_str)
                    pass
                    student_blocks_list = list(dict.fromkeys(student_blocks_list))
                    assert len(student_blocks_list) >= 1
                    student_blocks_list_list.append(student_blocks_list)
                pass
            pass  # end for student_block_type_list in seach_space_block_type_list_list:
        pass
        return student_blocks_list_list

class DeMainNet(PlainNet.PlainNet):
    def __init__(self, num_classes=None, plainnet_struct=None, no_create=False,
                 no_reslink=None, no_BN=None, use_se=None):

        super().__init__(num_classes=num_classes, plainnet_struct=plainnet_struct,
                                       no_create=no_create, no_reslink=no_reslink, no_BN=no_BN, use_se=use_se)
        self.last_channels = self.block_list[-1].out_channels
        self.fc_linear = basic_blocks.Linear(in_channels=self.last_channels, out_channels=self.num_classes, no_create=no_create)

        self.no_create = no_create
        self.no_reslink = no_reslink
        self.no_BN = no_BN
        self.use_se = use_se

        # bn eps
        for layer in self.modules():
            if isinstance(layer, nn.BatchNorm2d):
                layer.eps = 1e-3

    def extract_stage_features_and_logit(self, x, target_downsample_ratio=None):
        stage_features_list = []
        image_size = x.shape[2]
        output = x

        for block_id, the_block in enumerate(self.block_list):
            output = the_block(output)
            dowsample_ratio = round(image_size / output.shape[2])
            if dowsample_ratio == target_downsample_ratio:
                stage_features_list.append(output)
                target_downsample_ratio *= 2
            pass
        pass

        output = F.adaptive_avg_pool2d(output, output_size=1)
        output = torch.flatten(output, 1)
        logit = self.fc_linear(output)
        return stage_features_list, logit

    def forward(self, x):
        output = x
        for block_id, the_block in enumerate(self.block_list):
            output = the_block(output)

        output = F.adaptive_avg_pool2d(output, output_size=1)

        output = torch.flatten(output, 1)
        output = self.fc_linear(output)
        return output

    def forward_pre_GAP(self, x):
        output = x
        for the_block in self.block_list:
            output = the_block(output)
        return output

    def get_FLOPs(self, input_resolution):
        the_res = input_resolution
        the_flops = 0
        for the_block in self.block_list:
            the_flops += the_block.get_FLOPs(the_res)
            the_res = the_block.get_output_resolution(the_res)

        the_flops += self.fc_linear.get_FLOPs(the_res)

        return the_flops

    def get_model_size(self):
        the_size = 0
        for the_block in self.block_list:
            the_size += the_block.get_model_size()

        the_size += self.fc_linear.get_model_size()

        return the_size

    def get_num_layers(self):
        num_layers = 0
        for block in self.block_list:
            assert isinstance(block, super_blocks.PlainNetSuperBlockClass)
            num_layers += block.sub_layers
        return num_layers

    def replace_block(self, block_id, new_block):
        self.block_list[block_id] = new_block

        if block_id < len(self.block_list) - 1:
            if self.block_list[block_id + 1].in_channels != new_block.out_channels:
                self.block_list[block_id + 1].set_in_channels(new_block.out_channels)
        else:
            assert block_id == len(self.block_list) - 1
            self.last_channels = self.block_list[-1].out_channels
            if self.fc_linear.in_channels != self.last_channels:
                self.fc_linear.set_in_channels(self.last_channels)

        self.module_list = nn.ModuleList(self.block_list)

    def split(self, split_layer_threshold):
        new_str = ''
        for block in self.block_list:
            new_str += block.split(split_layer_threshold=split_layer_threshold)
        return new_str

    def init_parameters(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_normal_(m.weight.data, gain=3.26033)
                if hasattr(m, 'bias') and m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 3.26033 * np.sqrt(2 / (m.weight.shape[0] + m.weight.shape[1])))
                if hasattr(m, 'bias') and m.bias is not None:
                    nn.init.zeros_(m.bias)
            else:
                pass

        for superblock in self.block_list:
            if not isinstance(superblock, super_blocks.PlainNetSuperBlockClass):
                continue
            for block in superblock.block_list:
                if not (isinstance(block, basic_blocks.ResBlock) or isinstance(block, basic_blocks.ResBlockProj)):
                    continue
                # print('---debug set bn weight zero in resblock {}:{}'.format(superblock, block))
                last_bn_block = None
                for inner_resblock in block.block_list:
                    if isinstance(inner_resblock, basic_blocks.BN):
                        last_bn_block = inner_resblock
                    pass
                pass  # end for
                assert last_bn_block is not None
                # print('-------- last_bn_block={}'.format(last_bn_block))
                nn.init.zeros_(last_bn_block.netblock.weight)



from .cifar import DataBuilderCIFARMA
from e2eAIOK.common.trainer.data.cv.data_builder_uspsvsminist import DataBuilderUSPSMinist
import logging
import torch

# def channels_last_collate_distiller_train(batch):
#     # distiller : ((data, (logits, seed)), target)
#     data = [item[0][0] for item in batch]
#     data = torch.stack(data, 0).to(memory_format=torch.channels_last)

#     logits = [torch.from_numpy(item[0][1][0]) for item in batch]
#     logits = torch.stack(logits,0)

#     seed = [item[0][1][1] for item in batch]
#     seed = torch.Tensor(seed).type(torch.int32)

#     target = [item[1] for item in batch]
#     target = torch.LongTensor(target)

#     return [data, [logits, seed]], target

def createDataBuilder(cfg):
    if cfg.data_set == "USPS_vs_MNIST":
        dataBuilder = DataBuilderUSPSMinist(cfg=cfg)
        num_classes = 10
    elif cfg.data_set == "cifar10":
        dataBuilder = DataBuilderCIFARMA(cfg=cfg)
        num_classes = 10
    elif cfg.data_set == "cifar100":
        dataBuilder = DataBuilderCIFARMA(cfg=cfg)
        num_classes = 100
    else:
        raise NotImplementedError(f"Not support dataset {cfg.data_set}")
    return dataBuilder, num_classes





import math
import logging
import torch.nn as nn
from typing import Tuple

logger = logging.getLogger(__name__)


class Conv2d(nn.Module):
    """This function implements 2d convolution.

    Arguments
    ---------
    out_channels : int
        It is the number of output channels.
    kernel_size : tuple
        Kernel size of the 2d convolutional filters over time and frequency
        axis.
    input_shape : tuple
        The shape of the input. Alternatively use ``in_channels``.
    in_channels : int
        The number of input channels. Alternatively use ``input_shape``.
    stride: int
        Stride factor of the 2d convolutional filters over time and frequency
        axis.
    dilation : int
        Dilation factor of the 2d convolutional filters over time and
        frequency axis.
    padding : str
        (same, valid). If "valid", no padding is performed.
        If "same" and stride is 1, output shape is same as input shape.
    padding_mode : str
        This flag specifies the type of padding. See torch.nn documentation
        for more information.
    groups : int
        This option specifies the convolutional groups. See torch.nn
        documentation for more information.
    bias : bool
        If True, the additive bias b is adopted.
    skip_transpose : bool
        If False, uses batch x time x channel convention.
        If True, uses batch x channel x time convention.
    weight_norm : bool
        If True, use weight normalization,
        to be removed with self.remove_weight_norm() at inference

    Example
    -------
    >>> inp_tensor = torch.rand([10, 40, 16, 8])
    >>> cnn_2d = Conv2d(
    ...     input_shape=inp_tensor.shape, out_channels=5, kernel_size=(7, 3)
    ... )
    >>> out_tensor = cnn_2d(inp_tensor)
    >>> out_tensor.shape
    torch.Size([10, 40, 16, 5])
    """

    def __init__(
        self,
        out_channels,
        kernel_size,
        input_shape=None,
        in_channels=None,
        stride=(1, 1),
        dilation=(1, 1),
        padding="same",
        groups=1,
        bias=True,
        padding_mode="reflect",
        skip_transpose=False,
        weight_norm=False,
    ):
        super().__init__()

        # handle the case if some parameter is int
        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if isinstance(stride, int):
            stride = (stride, stride)
        if isinstance(dilation, int):
            dilation = (dilation, dilation)

        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.padding = padding
        self.padding_mode = padding_mode
        self.unsqueeze = False
        self.skip_transpose = skip_transpose

        if input_shape is None and in_channels is None:
            raise ValueError("Must provide one of input_shape or in_channels")

        if in_channels is None:
            in_channels = self._check_input(input_shape)

        self.in_channels = in_channels

        # Weights are initialized following pytorch approach
        self.conv = nn.Conv2d(
            self.in_channels,
            out_channels,
            self.kernel_size,
            stride=self.stride,
            padding=0,
            dilation=self.dilation,
            groups=groups,
            bias=bias,
        )

        if weight_norm:
            self.conv = nn.utils.weight_norm(self.conv)

    def forward(self, x):
        """Returns the output of the convolution.

        Arguments
        ---------
        x : torch.Tensor (batch, time, channel)
            input to convolve. 2d or 4d tensors are expected.

        """
        if not self.skip_transpose:
            x = x.transpose(1, -1)

        if self.unsqueeze:
            x = x.unsqueeze(1)

        if self.padding == "same":
            x = self._manage_padding(
                x, self.kernel_size, self.dilation, self.stride
            )

        elif self.padding == "valid":
            pass

        else:
            raise ValueError(
                "Padding must be 'same' or 'valid'. Got " + self.padding
            )

        wx = self.conv(x)

        if self.unsqueeze:
            wx = wx.squeeze(1)

        if not self.skip_transpose:
            wx = wx.transpose(1, -1)

        return wx

    def _manage_padding(
        self,
        x,
        kernel_size: Tuple[int, int],
        dilation: Tuple[int, int],
        stride: Tuple[int, int],
    ):
        """This function performs zero-padding on the time and frequency axes
        such that their lengths is unchanged after the convolution.

        Arguments
        ---------
        x : torch.Tensor
        kernel_size : int
        dilation : int
        stride: int
        """
        # Detecting input shape
        L_in = self.in_channels

        # Time padding
        padding_time = get_padding_elem(
            L_in, stride[-1], kernel_size[-1], dilation[-1]
        )

        padding_freq = get_padding_elem(
            L_in, stride[-2], kernel_size[-2], dilation[-2]
        )
        padding = padding_time + padding_freq

        # Applying padding
        x = nn.functional.pad(x, padding, mode=self.padding_mode)

        return x

    def _check_input(self, shape):
        """Checks the input shape and returns the number of input channels.
        """

        if len(shape) == 3:
            self.unsqueeze = True
            in_channels = 1

        elif len(shape) == 4:
            in_channels = shape[3]

        else:
            raise ValueError("Expected 3d or 4d inputs. Got " + len(shape))

        # Kernel size must be odd
        if self.kernel_size[0] % 2 == 0 or self.kernel_size[1] % 2 == 0:
            raise ValueError(
                "The field kernel size must be an odd number. Got %s."
                % (self.kernel_size)
            )

        return in_channels

    def remove_weight_norm(self):
        """Removes weight normalization at inference if used during training.
        """
        self.conv = nn.utils.remove_weight_norm(self.conv)


def get_padding_elem(L_in: int, stride: int, kernel_size: int, dilation: int):
    """This function computes the number of elements to add for zero-padding.

    Arguments
    ---------
    L_in : int
    stride: int
    kernel_size : int
    dilation : int
    """
    if stride > 1:
        padding = [math.floor(kernel_size / 2), math.floor(kernel_size / 2)]

    else:
        L_out = (
            math.floor((L_in - dilation * (kernel_size - 1) - 1) / stride) + 1
        )
        padding = [
            math.floor((L_in - L_out) / 2),
            math.floor((L_in - L_out) / 2),
        ]
    return padding



#    Copyright 2022, Intel Corporation.
#    Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from torch import nn
import torch
import numpy as np
import torch.nn.functional
from nnunet.network_architecture.initialization import InitWeights_He
from nnunet.network_architecture.neural_network import SegmentationNetwork
from nnunet.network_architecture.generic_UNet import ConvDropoutNormNonlin,StackedConvLayers,Upsample
from nnunet.utilities.nd_softmax import softmax_helper


def get_prefixed_named_param(model, prefix):
    for name, param in model.named_parameters():
        if name.startswith(prefix):
            yield param

def get_unprefixed_named_param(model, prefix):
    for name, param in model.named_parameters():
        if not name.startswith(prefix):
            yield param


class Generic_UNet(SegmentationNetwork):
    DEFAULT_BATCH_SIZE_3D = 2
    DEFAULT_PATCH_SIZE_3D = (64, 192, 160)
    SPACING_FACTOR_BETWEEN_STAGES = 2
    BASE_NUM_FEATURES_3D = 30
    MAX_NUMPOOL_3D = 999
    MAX_NUM_FILTERS_3D = 320

    DEFAULT_PATCH_SIZE_2D = (256, 256)
    BASE_NUM_FEATURES_2D = 30
    DEFAULT_BATCH_SIZE_2D = 50
    MAX_NUMPOOL_2D = 999
    MAX_FILTERS_2D = 480

    use_this_for_batch_size_computation_2D = 19739648
    use_this_for_batch_size_computation_3D = 520000000  # 505789440

    def __init__(self, input_channels, base_num_features, num_classes, num_pool, num_conv_per_stage=2,
                 feat_map_mul_on_downscale=2, conv_op=nn.Conv2d,
                 norm_op=nn.BatchNorm2d, norm_op_kwargs=None,
                 dropout_op=nn.Dropout2d, dropout_op_kwargs=None,
                 nonlin=nn.LeakyReLU, nonlin_kwargs=None, deep_supervision=True, dropout_in_localization=False,
                 final_nonlin=softmax_helper, weightInitializer=InitWeights_He(1e-2), pool_op_kernel_sizes=None,
                 conv_kernel_sizes=None,
                 upscale_logits=False, convolutional_pooling=False, convolutional_upsampling=False,
                 max_num_features=None, basic_block=ConvDropoutNormNonlin,
                 seg_output_use_bias=False):
        """
        basically more flexible than v1, architecture is the same

        Does this look complicated? Nah bro. Functionality > usability

        This does everything you need, including world peace.

        Questions? -> f.isensee@dkfz.de
        """
        super(Generic_UNet, self).__init__()
        self.convolutional_upsampling = convolutional_upsampling
        self.convolutional_pooling = convolutional_pooling
        self.upscale_logits = upscale_logits
        if nonlin_kwargs is None:
            nonlin_kwargs = {'negative_slope': 1e-2, 'inplace': True}
        if dropout_op_kwargs is None:
            dropout_op_kwargs = {'p': 0.5, 'inplace': True}
        if norm_op_kwargs is None:
            norm_op_kwargs = {'eps': 1e-5, 'affine': True, 'momentum': 0.1}

        self.conv_kwargs = {'stride': 1, 'dilation': 1, 'bias': True}

        self.nonlin = nonlin
        self.nonlin_kwargs = nonlin_kwargs
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.weightInitializer = weightInitializer
        self.conv_op = conv_op
        self.norm_op = norm_op
        self.dropout_op = dropout_op
        self.num_classes = num_classes
        self.final_nonlin = final_nonlin
        self._deep_supervision = deep_supervision

        if conv_op == nn.Conv2d:
            upsample_mode = 'bilinear'
            pool_op = nn.MaxPool2d
            transpconv = nn.ConvTranspose2d
            if pool_op_kernel_sizes is None:
                pool_op_kernel_sizes = [(2, 2)] * num_pool
            if conv_kernel_sizes is None:
                conv_kernel_sizes = [(3, 3)] * (num_pool + 1)
        elif conv_op == nn.Conv3d:
            upsample_mode = 'trilinear'
            pool_op = nn.MaxPool3d
            transpconv = nn.ConvTranspose3d
            if pool_op_kernel_sizes is None:
                pool_op_kernel_sizes = [(2, 2, 2)] * num_pool
            if conv_kernel_sizes is None:
                conv_kernel_sizes = [(3, 3, 3)] * (num_pool + 1)
        else:
            raise ValueError("unknown convolution dimensionality, conv op: %s" % str(conv_op))

        self.input_shape_must_be_divisible_by = np.prod(pool_op_kernel_sizes, 0, dtype=np.int64)
        self.pool_op_kernel_sizes = pool_op_kernel_sizes
        self.conv_kernel_sizes = conv_kernel_sizes

        self.conv_pad_sizes = []
        for krnl in self.conv_kernel_sizes:
            self.conv_pad_sizes.append([1 if i == 3 else 0 for i in krnl])

        if max_num_features is None:
            if self.conv_op == nn.Conv3d:
                self.max_num_features = self.MAX_NUM_FILTERS_3D
            else:
                self.max_num_features = self.MAX_FILTERS_2D
        else:
            self.max_num_features = max_num_features

        self.conv_blocks_context = []
        self.conv_blocks_localization = []
        self.td = []
        self.tu = []
        self.seg_outputs = []

        output_features = base_num_features
        input_features = input_channels

        self.encoder_channels = []
        for d in range(num_pool):
            # determine the first stride
            if d != 0 and self.convolutional_pooling:
                first_stride = pool_op_kernel_sizes[d - 1]
            else:
                first_stride = None

            self.conv_kwargs['kernel_size'] = self.conv_kernel_sizes[d]
            self.conv_kwargs['padding'] = self.conv_pad_sizes[d]
            # add convolutions
            self.conv_blocks_context.append(StackedConvLayers(input_features, output_features, num_conv_per_stage,
                                                              self.conv_op, self.conv_kwargs, self.norm_op,
                                                              self.norm_op_kwargs, self.dropout_op,
                                                              self.dropout_op_kwargs, self.nonlin, self.nonlin_kwargs,
                                                              first_stride, basic_block=basic_block))
            self.encoder_channels.append(output_features)
            if not self.convolutional_pooling:
                self.td.append(pool_op(pool_op_kernel_sizes[d]))
            input_features = output_features
            output_features = int(np.round(output_features * feat_map_mul_on_downscale))

            output_features = min(output_features, self.max_num_features)

        # now the bottleneck.
        # determine the first stride
        if self.convolutional_pooling:
            first_stride = pool_op_kernel_sizes[-1]
        else:
            first_stride = None

        # the output of the last conv must match the number of features from the skip connection if we are not using
        # convolutional upsampling. If we use convolutional upsampling then the reduction in feature maps will be
        # done by the transposed conv
        if self.convolutional_upsampling:
            final_num_features = output_features
        else:
            final_num_features = self.conv_blocks_context[-1].output_channels

        self.conv_kwargs['kernel_size'] = self.conv_kernel_sizes[num_pool]
        self.conv_kwargs['padding'] = self.conv_pad_sizes[num_pool]
        self.conv_blocks_context.append(nn.Sequential(
            StackedConvLayers(input_features, output_features, num_conv_per_stage - 1, self.conv_op, self.conv_kwargs,
                              self.norm_op, self.norm_op_kwargs, self.dropout_op, self.dropout_op_kwargs, self.nonlin,
                              self.nonlin_kwargs, first_stride, basic_block=basic_block),
            StackedConvLayers(output_features, final_num_features, 1, self.conv_op, self.conv_kwargs,
                              self.norm_op, self.norm_op_kwargs, self.dropout_op, self.dropout_op_kwargs, self.nonlin,
                              self.nonlin_kwargs, basic_block=basic_block)))

        # if we don't want to do dropout in the localization pathway then we set the dropout prob to zero here
        if not dropout_in_localization:
            old_dropout_p = self.dropout_op_kwargs['p']
            self.dropout_op_kwargs['p'] = 0.0

        # now lets build the localization pathway
        for u in range(num_pool):
            nfeatures_from_down = final_num_features
            nfeatures_from_skip = self.conv_blocks_context[
                -(2 + u)].output_channels  # self.conv_blocks_context[-1] is bottleneck, so start with -2
            n_features_after_tu_and_concat = nfeatures_from_skip * 2

            # the first conv reduces the number of features to match those of skip
            # the following convs work on that number of features
            # if not convolutional upsampling then the final conv reduces the num of features again
            if u != num_pool - 1 and not self.convolutional_upsampling:
                final_num_features = self.conv_blocks_context[-(3 + u)].output_channels
            else:
                final_num_features = nfeatures_from_skip

            if not self.convolutional_upsampling:
                self.tu.append(Upsample(scale_factor=pool_op_kernel_sizes[-(u + 1)], mode=upsample_mode))
            else:
                self.tu.append(transpconv(nfeatures_from_down, nfeatures_from_skip, pool_op_kernel_sizes[-(u + 1)],
                                          pool_op_kernel_sizes[-(u + 1)], bias=False))

            self.conv_kwargs['kernel_size'] = self.conv_kernel_sizes[- (u + 1)]
            self.conv_kwargs['padding'] = self.conv_pad_sizes[- (u + 1)]
            self.conv_blocks_localization.append(nn.Sequential(
                StackedConvLayers(n_features_after_tu_and_concat, nfeatures_from_skip, num_conv_per_stage - 1,
                                  self.conv_op, self.conv_kwargs, self.norm_op, self.norm_op_kwargs, self.dropout_op,
                                  self.dropout_op_kwargs, self.nonlin, self.nonlin_kwargs, basic_block=basic_block),
                StackedConvLayers(nfeatures_from_skip, final_num_features, 1, self.conv_op, self.conv_kwargs,
                                  self.norm_op, self.norm_op_kwargs, self.dropout_op, self.dropout_op_kwargs,
                                  self.nonlin, self.nonlin_kwargs, basic_block=basic_block)
            ))

        for ds in range(len(self.conv_blocks_localization)):
            self.seg_outputs.append(conv_op(self.conv_blocks_localization[ds][-1].output_channels, num_classes,
                                            1, 1, 0, 1, 1, seg_output_use_bias))

        self.upscale_logits_ops = []
        cum_upsample = np.cumprod(np.vstack(pool_op_kernel_sizes), axis=0)[::-1]
        for usl in range(num_pool - 1):
            if self.upscale_logits:
                self.upscale_logits_ops.append(Upsample(scale_factor=tuple([int(i) for i in cum_upsample[usl + 1]]),
                                                        mode=upsample_mode))
            else:
                self.upscale_logits_ops.append(lambda x: x)

        if not dropout_in_localization:
            self.dropout_op_kwargs['p'] = old_dropout_p

        # register all modules properly
        self.conv_blocks_localization = nn.ModuleList(self.conv_blocks_localization)
        self.conv_blocks_context = nn.ModuleList(self.conv_blocks_context)
        self.td = nn.ModuleList(self.td)
        self.tu = nn.ModuleList(self.tu)
        self.seg_outputs = nn.ModuleList(self.seg_outputs)
        if self.upscale_logits:
            self.upscale_logits_ops = nn.ModuleList(
                self.upscale_logits_ops)  # lambda x:x is not a Module so we need to distinguish here

        if self.weightInitializer is not None:
            self.apply(self.weightInitializer)
            # self.apply(print_module_training_status)

    def forward(self, x):
        skips = []
        seg_outputs = []
        for d in range(len(self.conv_blocks_context) - 1):
            x = self.conv_blocks_context[d](x)
            skips.append(x)
            if not self.convolutional_pooling:
                x = self.td[d](x)

        x = self.conv_blocks_context[-1](x)

        for u in range(len(self.tu)):
            x = self.tu[u](x)
            x = torch.cat((x, skips[-(u + 1)]), dim=1)
            x = self.conv_blocks_localization[u](x)
            seg_outputs.append(self.final_nonlin(self.seg_outputs[u](x)))

        if self._deep_supervision:
            logits = tuple([seg_outputs[-1]] + [i(j) for i, j in
                                              zip(list(self.upscale_logits_ops)[::-1], seg_outputs[:-1][::-1])])
        else:
            logits = tuple([seg_outputs[-1]])
            
        if self.training:
            return logits
        else:
            return logits[0]

    def get_parameters(self, model_finetune=True):
        parameters = []
        if model_finetune:
            parameters += [{
                'params': get_unprefixed_named_param(self, 'seg_outputs')
            }]
            parameters += [{
                'params': get_prefixed_named_param(self, 'seg_outputs')
            }]
        else:
            parameters = [{'params': self.parameters()}]
        return parameters

    @staticmethod
    def compute_approx_vram_consumption(patch_size, num_pool_per_axis, base_num_features, max_num_features,
                                        num_modalities, num_classes, pool_op_kernel_sizes, deep_supervision=False,
                                        conv_per_stage=2):
        """
        This only applies for num_conv_per_stage and convolutional_upsampling=True
        not real vram consumption. just a constant term to which the vram consumption will be approx proportional
        (+ offset for parameter storage)
        :param deep_supervision:
        :param patch_size:
        :param num_pool_per_axis:
        :param base_num_features:
        :param max_num_features:
        :param num_modalities:
        :param num_classes:
        :param pool_op_kernel_sizes:
        :return:
        """
        if not isinstance(num_pool_per_axis, np.ndarray):
            num_pool_per_axis = np.array(num_pool_per_axis)

        npool = len(pool_op_kernel_sizes)

        map_size = np.array(patch_size)
        tmp = np.int64((conv_per_stage * 2 + 1) * np.prod(map_size, dtype=np.int64) * base_num_features +
                       num_modalities * np.prod(map_size, dtype=np.int64) +
                       num_classes * np.prod(map_size, dtype=np.int64))

        num_feat = base_num_features

        for p in range(npool):
            for pi in range(len(num_pool_per_axis)):
                map_size[pi] /= pool_op_kernel_sizes[p][pi]
            num_feat = min(num_feat * 2, max_num_features)
            num_blocks = (conv_per_stage * 2 + 1) if p < (npool - 1) else conv_per_stage  # conv_per_stage + conv_per_stage for the convs of encode/decode and 1 for transposed conv
            tmp += num_blocks * np.prod(map_size, dtype=np.int64) * num_feat
            if deep_supervision and p < (npool - 2):
                tmp += np.prod(map_size, dtype=np.int64) * num_classes
            # print(p, map_size, num_feat, tmp)
        return tmp



import subprocess
import yaml
import logging
import os

from e2eAIOK.common.utils import *
from e2eAIOK.SDA.modeladvisor.BaseModelAdvisor import BaseModelAdvisor

class BERTAdvisor(BaseModelAdvisor):
    '''
        Bert sigopt model optimization launcher
    '''
    def __init__(self, dataset_meta_path, train_path, eval_path, args):
        super().__init__(dataset_meta_path, train_path, eval_path, args)
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('sigopt')
        self.python_path = "/opt/intel/oneapi/intelpython/latest/envs/tensorflow/bin/"
        self.train_python = f"{self.python_path}/python"
        self.train_script = os.path.join(os.getcwd(),"modelzoo/bert/benchmarks/launch_benchmark.py")
        self.train_path = train_path
        self.test_path = eval_path
        self.dataset_meta_path = dataset_meta_path
        self.current_path = os.getcwd()
        self.extra_pythonpath = f"{self.current_path}/modelzoo/bert/benchmarks/"

    def initialize_model_parameter(self, assignments=None):
        '''
            Add model specific parameters
            For sigopt parameters, set parameter explicitly and the parameter will not be optimized by sigopt
        '''
        config = {}
        tuned_parameters = {}
        if assignments:
            tuned_parameters["train_batch_size"] = str(assignments['train_batch_size'])
            tuned_parameters["learning_rate"] = str(assignments['learning_rate'])
            tuned_parameters["warmup_proportion"] = str(assignments['warmup_proportion'])
            tuned_parameters["num_hidden_layers"] = str(assignments['num_hidden_layers'])
            if 'dropout_prob' in assignments:
                tuned_parameters["attention_probs_dropout_prob"] = str(assignments['dropout_prob'])
                tuned_parameters["hidden_dropout_prob"] = str(assignments['dropout_prob'])
            else:
                tuned_parameters["attention_probs_dropout_prob"] = str(assignments['attention_probs_dropout_prob'])
                tuned_parameters["hidden_dropout_prob"] = str(assignments['hidden_dropout_prob'])
        else:
            tuned_parameters["train_batch_size"] = "24"
            tuned_parameters["learning_rate"] = "3e-5"
            tuned_parameters["warmup_proportion"] = "0.1"
            tuned_parameters["num_hidden_layers"] = "24"
            tuned_parameters["attention_probs_dropout_prob"] = "0.1"
            tuned_parameters["hidden_dropout_prob"] = "0.1"
        config['tuned_parameters'] = tuned_parameters
        self.params['model_parameter'] = config
        self.params['model_saved_path'] = os.path.join(
            self.params['save_path'], get_hash_string(str(tuned_parameters)))
        return config

    ###### Implementation of required methods ######
    
    def generate_sigopt_yaml(self, file='bert_sigopt.yaml'):
        config = {}
        config['project'] = 'bert'
        config['experiment'] = 'BERT'
        parameters = [
        {'name':'learning_rate','bounds':{'min':1e-6, 'max':3e-4},'type':'double'},
        {'name':'train_batch_size','grid':[16,24,32,64,96],'type':'int'},
        {'name':'num_hidden_layers', 'grid':[8,12,16,20,24], 'type':'int'},
        {'name':'dropout_prob', 'grid':[0.08,0.1,0.12], 'type':'double'},
        {'name':'warmup_proportion', 'bounds':{'min':0.05, 'max':0.20},'type':'double'},
        ]
        config['parameters'] = parameters
        metrics = []
        if 'training_time_threshold' in self.params:
            metrics.append({'name': 'training_time', 'objective': 'minimize', 'threshold': self.params['training_time_threshold']})
        else:
            metrics.append({'name': 'training_time', 'objective': 'minimize'})
        if 'metric_threshold' in self.params:
            metrics.append({'name': self.params['metric'], 'objective': self.params['metric_objective'], 'threshold': self.params['metric_threshold']})
        else:
            metrics.append({'name': self.params['metric'], 'objective': self.params['metric_objective']})
        
        config['metrics'] = metrics
        config['observation_budget'] = self.params['observation_budget']

        with open(file, 'w') as f:
            yaml.dump(config, f)
        return config

    def update_metrics(self):
        metrics = []
        metrics.append({'name': 'f1', 'value': self.mean_accuracy})
        metrics.append({'name': 'training_time', 'value': self.training_time})
        self.params['model_parameter']['metrics'] = metrics
        return self.params['model_parameter']['metrics']

    def train_model(self, args):
        model_path = args['model_saved_path']
        if int(args['mpi_num_processes']) > 1:
            self.dist_launch(args)
        else:
            self.launch(args)
        with open(os.path.join(self.params['model_saved_path'],"best_auc.txt"),'r') as f:
            lines = f.readlines()
        self.mean_accuracy = float(lines[-1])
        with open(os.path.join(self.params['model_saved_path'],"best_time.txt"),'r') as f:
            lines = f.readlines()
        self.training_time = float(lines[-1])
        metrics = self.update_metrics()
        return self.training_time, model_path, metrics

    def dist_launch(self, args):
        # construct BERT launch command
        os.environ["MODEL_DIR"] = MODEL_DIR = os.path.join(os.getcwd(),"modelzoo/bert")
        os.environ["OUTPUT_DIR"] = OUT_DIR = self.params['model_saved_path']
        os.environ["DATASET_DIR"] = os.path.dirname(self.dataset_meta_path)
        os.environ["CHECKPOINT_DIR"] = "$DATASET_DIR/pre-trained-model/bert-large-uncased/wwm_uncased_L-24_H-1024_A-16"
        os.environ["HOROVOD_CPU_OPERATIONS"] = "CCL"
        os.environ["NOINSTALL"] = "True"
        os.environ["CCL_WORKER_COUNT"] = str(args['ccl_worker_num'])
        os.environ["CCL_WORKER_AFFINITY"] = "16,17,34,35"
        os.environ["HOROVOD_THREAD_AFFINITY"] = "53,71"
        os.environ["I_MPI_PIN_DOMAIN"] = "socket"
        os.environ["I_MPI_PIN_PROCESSOR_EXCLUDE_LIST"] = "16,17,34,35,52,53,70,71"
        mpi_num_proc_arg = ""

        if not os.path.exists(os.environ["OUTPUT_DIR"]):
            os.makedirs(os.environ["OUTPUT_DIR"])

        cmd = f"PYTHONPATH=$PYTHONPATH:{self.extra_pythonpath} "
        cmd += f"{self.train_python} "
        cmd += f"{MODEL_DIR}/benchmarks/launch_benchmark.py " \
        + f"--model-name=bert_large " \
        + f"--precision=fp32 " \
        + f"--mode=training " \
        + f"--framework=tensorflow {mpi_num_proc_arg}" \
        + f"--batch-size={args['model_parameter']['tuned_parameters']['train_batch_size']} " \
        + f"--output-dir {OUT_DIR} " \
        + f"--host_file=$MODEL_DIR/hosts " \
        + f"--mpi_num_processes {args['mpi_num_processes']} " \
        + f"--num-intra-threads 36 " \
        + f"--num-inter-threads 2 " \
        + f"--train_option=SQuAD " \
        + f"vocab_file=$CHECKPOINT_DIR/vocab.txt " \
        + f"config_file=$CHECKPOINT_DIR/bert_config.json " \
        + f"init_checkpoint=$CHECKPOINT_DIR/bert_model.ckpt " \
        + f"do_train=True " \
        + f"train_file=$DATASET_DIR/train-v1.1.json " \
        + f"do_predict=True " \
        + f"predict_file=$DATASET_DIR/dev-v1.1.json " \
        + f"test_file=$DATASET_DIR/test-v1.1.json " \
        + f"data_dir=$DATASET_DIR " \
        + f"num_to_evaluate=50 " \
        + f"step_threshold={args['step_threshold']} " \
        + f"f1_threshold=90.87 " \
        + f"num_train_epochs=2 " \
        + f"max_seq_length=384 " \
        + f"doc_stride=128 " \
        + f"optimized_softmax=True " \
        + f"experimental_gelu=False " \
        + f"do_lower_case=True " \
        + f"num_hidden_layers={args['model_parameter']['tuned_parameters']['num_hidden_layers']} " \
        + f"learning_rate={args['model_parameter']['tuned_parameters']['learning_rate']} " \
        + f"attention_probs_dropout_prob={args['model_parameter']['tuned_parameters']['attention_probs_dropout_prob']} " \
        + f"hidden_dropout_prob={args['model_parameter']['tuned_parameters']['hidden_dropout_prob']} " \
        + f"warmup_proportion={args['model_parameter']['tuned_parameters']['warmup_proportion']} "

        print(f'training launch command: {cmd}')
        process = subprocess.Popen(cmd, shell=True)
        process.wait()

    def launch(self, args):
        # construct BERT launch command
        os.environ["MODEL_DIR"] = MODEL_DIR = os.path.join(os.getcwd(),"modelzoo/bert")
        os.environ["OUTPUT_DIR"] = OUT_DIR = self.params['model_saved_path']
        os.environ["DATASET_DIR"] = os.path.dirname(self.dataset_meta_path)
        os.environ["CHECKPOINT_DIR"] = "$DATASET_DIR/pre-trained-model/bert-large-uncased/wwm_uncased_L-24_H-1024_A-16"
        mpi_num_proc_arg = ""

        if not os.path.exists(os.environ["OUTPUT_DIR"]):
            os.makedirs(os.environ["OUTPUT_DIR"])

        cmd = f"PYTHONPATH=$PYTHONPATH:{self.extra_pythonpath} "
        cmd += f"{self.train_python} "
        cmd += f"{MODEL_DIR}/benchmarks/launch_benchmark.py " \
        + f"--model-name=bert_large " \
        + f"--precision=fp32 " \
        + f"--mode=training " \
        + f"--framework=tensorflow {mpi_num_proc_arg}" \
        + f"--batch-size={args['model_parameter']['tuned_parameters']['train_batch_size']} " \
        + f"--output-dir {OUT_DIR} " \
        + f"--host_file=$MODEL_DIR/hosts " \
        + f"--num-intra-threads 36 " \
        + f"--num-inter-threads 2 " \
        + f"--train_option=SQuAD " \
        + f"vocab_file=$CHECKPOINT_DIR/vocab.txt " \
        + f"config_file=$CHECKPOINT_DIR/bert_config.json " \
        + f"init_checkpoint=$CHECKPOINT_DIR/bert_model.ckpt " \
        + f"do_train=True " \
        + f"train_file=$DATASET_DIR/train-v1.1.json " \
        + f"do_predict=True " \
        + f"predict_file=$DATASET_DIR/dev-v1.1.json " \
        + f"test_file=$DATASET_DIR/test-v1.1.json " \
        + f"data_dir=$DATASET_DIR " \
        + f"num_to_evaluate=50 " \
        + f"step_threshold={args['step_threshold']} " \
        + f"f1_threshold=90.87 " \
        + f"num_train_epochs=2 " \
        + f"max_seq_length=384 " \
        + f"doc_stride=128 " \
        + f"optimized_softmax=True " \
        + f"experimental_gelu=False " \
        + f"do_lower_case=True " \
        + f"num_hidden_layers={args['model_parameter']['tuned_parameters']['num_hidden_layers']} " \
        + f"learning_rate={args['model_parameter']['tuned_parameters']['learning_rate']} " \
        + f"attention_probs_dropout_prob={args['model_parameter']['tuned_parameters']['attention_probs_dropout_prob']} " \
        + f"hidden_dropout_prob={args['model_parameter']['tuned_parameters']['hidden_dropout_prob']} " \
        + f"warmup_proportion={args['model_parameter']['tuned_parameters']['warmup_proportion']} "

        print(f'training launch command: {cmd}')
        process = subprocess.Popen(cmd, shell=True)
        process.wait()



import torch
import collections
import collections.abc
import os
import time
import yaml
import pathlib
import inspect
import shutil
import logging
import warnings

logger = logging.getLogger(__name__)

CKPT_PREFIX = "CKPT"
METAFNAME = f"{CKPT_PREFIX}.yaml"  # Important that this is not .ckpt
PARAMFILE_EXT = ".ckpt"  # ...because these files will be


def torch_recovery(obj, path, end_of_epoch, device=None):
    """Loads a torch.nn.Module state_dict from the given path instantly.

    This can be made the default for torch.nn.Modules with:
    >>> DEFAULT_LOAD_HOOKS[torch.nn.Module] = torch_recovery

    Arguments
    ---------
    obj : torch.nn.Module
        Instance for which to load the parameters.
    path : str, pathlib.Path
        Path where to load from.
    end_of_epoch : bool
        Whether the recovery comes from an end of epoch checkpoint.
    device : str
        Torch device, where to map the loaded parameters.

    Returns
    -------
    None
        Given object is modified in place.
    """
    del end_of_epoch  # Unused
    try:
        obj.load_state_dict(torch.load(path, map_location=device), strict=True)
    except TypeError:
        obj.load_state_dict(torch.load(path, map_location=device))


def torch_save(obj, path):
    """Saves the obj's parameters to path.

    Default save hook for torch.nn.Modules
    For saving torch.nn.Module state_dicts.

    Arguments
    ---------
    obj : torch.nn.Module
        Instance to save.
    path : str, pathlib.Path
        Path where to save to.

    Returns
    -------
    None
        State dict is written to disk.
    """
    state_dict = obj.state_dict()
    if not state_dict:
        logger.warning(f"Saving an empty state_dict for {obj} in {path}.")
    torch.save(state_dict, path)


def torch_parameter_transfer(obj, path, device):
    """Non-strict Torch Module state_dict load.

    Loads a set of parameters from path to obj. If obj has layers for which
    parameters can't be found, only a warning is logged. Same thing
    if the path has parameters for layers which don't find a counterpart
    in obj.

    Arguments
    ---------
    obj : torch.nn.Module
        Instance for which to load the parameters.
    path : str
        Path where to load from.

    Returns
    -------
    None
        The object is modified in place.
    """
    incompatible_keys = obj.load_state_dict(
        torch.load(path, map_location=device), strict=False
    )
    for missing_key in incompatible_keys.missing_keys:
        logger.warning(
            f"During parameter transfer to {obj} loading from "
            + f"{path}, the transferred parameters did not have "
            + f"parameters for the key: {missing_key}"
        )
    for unexpected_key in incompatible_keys.unexpected_keys:
        logger.warning(
            f"During parameter transfer to {obj} loading from "
            + f"{path}, the object could not use the parameters loaded "
            + f"with the key: {unexpected_key}"
        )


# These dicts are indexed by class and hold the default checkpoints methods
DEFAULT_LOAD_HOOKS = {
    torch.nn.Module: torch_recovery,
    torch.optim.Optimizer: torch_recovery,
    torch.optim.lr_scheduler._LRScheduler: torch_recovery,
    torch.optim.lr_scheduler.ReduceLROnPlateau: torch_recovery,
}
DEFAULT_SAVE_HOOKS = {
    torch.nn.Module: torch_save,
    torch.optim.Optimizer: torch_save,
    torch.optim.lr_scheduler._LRScheduler: torch_save,
    torch.optim.lr_scheduler.ReduceLROnPlateau: torch_save,
}
DEFAULT_TRANSFER_HOOKS = {
    torch.nn.Module: torch_parameter_transfer,
}

# Add a transfer hook for sentencepiece if it is installed:
try:
    import sentencepiece as spm

    def _load_spm(obj, path, device=None):
        obj.load(str(path))  # SentencePieceProcessor needs a string.

    DEFAULT_TRANSFER_HOOKS[spm.SentencePieceProcessor] = _load_spm
    del spm  # Don't leave it here bare.
except ImportError:
    # SentencePiece not loaded, fine!
    pass


def mark_as_saver(method):
    """Method decorator which marks given method as the checkpoint saving hook.

    See register_checkpoint_hooks for example.

    Arguments
    ---------
    method : callable
        Method of the class to decorate. Must be callable with
        signature (instance, path) using positional arguments. This is
        satisfied by for example: def saver(self, path):

    Note
    ----
    This will not add the hook (not possible via a method decorator),
    you must also decorate the class with @register_checkpoint_hooks
    Only one method can be added as the hook.
    """
    sig = inspect.signature(method)
    try:
        sig.bind(object(), pathlib.Path("testpath"))
    except TypeError:
        MSG = "Checkpoint saver must match signature (instance, path)"
        raise TypeError(MSG)
    method._speechbrain_saver = True
    return method


def mark_as_loader(method):
    """Method decorator which marks given method as checkpoint loading hook.

    Arguments
    ---------
    method : callable
        Method of the class to decorate. Must be callable with
        signature (instance, path, end_of_epoch, device) using positional
        arguments. This is satisfied by for example:
        `def loader(self, path, end_of_epoch, device):`

    Note
    ----
    This will not add the hook (not possible via a method decorator),
    you must also decorate the class with @register_checkpoint_hooks
    Only one method can be added as the hook.
    """
    sig = inspect.signature(method)
    try:
        sig.bind(object(), pathlib.Path("testpath"), True, None)
    except TypeError:
        MSG = "Checkpoint loader must have signature (self, path, end_of_epoch, device)"
        raise TypeError(MSG)
    method._speechbrain_loader = True
    return method


def mark_as_transfer(method):
    """Method decorator which marks given method as a parameter transfer hook.

    Arguments
    ---------
    method : callable
        Method of the class to decorate. Must be callable with
        signature (instance, path, device) using positional
        arguments. This is satisfied by for example:
        `def loader(self, path, device):`

    Note
    ----
    This will not add the hook (not possible via a method decorator),
    you must also decorate the class with @register_checkpoint_hooks
    Only one method can be added as the hook.

    Note
    ----
    The transfer hook is prioritized over the loader hook by the ``Pretrainer``
    However, if no transfer hook is registered, the Pretrainer will use the
    loader hook.
    """
    sig = inspect.signature(method)
    try:
        sig.bind(object(), pathlib.Path("testpath"), device=None)
    except TypeError:
        MSG = "Transfer hook must have signature (self, path, device)"
        raise TypeError(MSG)
    method._speechbrain_transfer = True
    return method


def register_checkpoint_hooks(cls):
    """Class decorator which registers the load, save and transfer hooks.

    The hooks must have been marked with mark_as_loader and mark_as_saver,
    and possibly mark_as_transfer.

    Arguments
    ---------
    cls : class
        Class to decorate
    """
    global DEFAULT_LOAD_HOOKS
    global DEFAULT_SAVE_HOOKS
    global DEFAULT_TRANSFER_HOOKS
    for name, method in cls.__dict__.items():
        if hasattr(method, "_speechbrain_saver"):
            DEFAULT_SAVE_HOOKS[cls] = method
            logger.debug(f"Registered checkpoint save hook for {name}")
        if hasattr(method, "_speechbrain_loader"):
            DEFAULT_LOAD_HOOKS[cls] = method
            logger.debug(f"Registered checkpoint load hook for {name}")
        if hasattr(method, "_speechbrain_transfer"):
            DEFAULT_TRANSFER_HOOKS[cls] = method
            logger.debug(f"Registered parameter transfer hook for {name}")
    return cls


def get_default_hook(obj, default_hooks):
    """Finds the default save/load hook to use with the given object.

    Follows the Method Resolution Order, i.e., if no hook is registered for
    the class of the object itself, also searches classes which the object
    inherits from.

    Arguments
    ---------
    obj : instance
        Instance of a class.
    default_hooks : dict
        Mapping from classes to (checkpointing hook) functions.

    Returns
    -------
    The correct method or None if no method is registered.
    """
    mro = inspect.getmro(type(obj))
    for cls in mro:
        if cls in default_hooks:
            return default_hooks[cls]
    # If we got here, no hook found
    return None


Checkpoint = collections.namedtuple(
    "Checkpoint", ["path", "meta", "paramfiles"]
)
Checkpoint.__doc__ = """NamedTuple describing one saved checkpoint

To select a checkpoint to load from many checkpoint,
Checkpoints are first filtered and sorted based on this namedtuple.
Checkpointers put pathlib.Path in path and a dict in meta.
You can essentially add any info you want to meta when saving a checkpoint.
The only default key in meta is "unixtime".
Checkpoint.paramfiles is a dict from recoverable name to parameter filepath.
"""
# Creating a hash allows making checkpoint sets
Checkpoint.__hash__ = lambda self: hash(self.path)


def ckpt_recency(ckpt):
    """Recency as Checkpoint importance metric.

    This function can also act as an example of how to make checkpoint
    importance keyfuncs. This is a named function, but as you can see
    it could be easily implemented as a lambda in a pinch.
    """
    return ckpt.meta["unixtime"]


class Checkpointer:
    """Saves checkpoints and recovers from them.

    Arguments:

    checkpoints_dir : str, pathlib.Path
        Path to directory where to save checkpoints.
    recoverables : mapping, optional
        Objects to to recover. They need a (unique) name: this is used
        to connect the parameters in a checkpoint to the correct recoverable.
        The name is also used in the filename of the
        savefile for the objects parameters. These can also be added with
        add_recoverable or add_recoverables or just modifying
        checkpointer.recoverables directly.
    custom_load_hooks : mapping, optional
        A mapping from name [same as in recoverables] to function or method.
        Sets a custom loading hook for a particular object. The
        function/method must be callable with signature (instance, path)
        using positional arguments. This is satisfied by for example:
        `def loader(self, path)`.
    custom_save_hooks : mapping, optional
        Mapping from name [same as in recoverables] to function or method.
        Sets a custom saving hook for a particular object. The
        function/method must be callable with
        signature (instance, path) using positional arguments. This is
        satisfied by for example: def saver(self, path):
    allow_partial_load : bool, optional
        If True, allows loading a checkpoint where a savefile is not found
        for every registered recoverable. In that case, only the found
        savefiles are loaded. When False, loading such a save will raise
        RuntimeError. (default: False)
    """

    def __init__(
        self,
        checkpoints_dir,
        recoverables=None,
        custom_load_hooks=None,
        custom_save_hooks=None,
        allow_partial_load=False,
    ):
        self.checkpoints_dir = pathlib.Path(checkpoints_dir)
        os.makedirs(self.checkpoints_dir, exist_ok=True)
        self.recoverables = {}
        if recoverables is not None:
            self.add_recoverables(recoverables)
        self.custom_load_hooks = {}
        if custom_load_hooks is not None:
            self.custom_load_hooks.update(custom_load_hooks)
        self.custom_save_hooks = {}
        if custom_save_hooks is not None:
            self.custom_save_hooks.update(custom_save_hooks)
        self.allow_partial_load = allow_partial_load

    def add_recoverable(
        self, name, obj, custom_load_hook=None, custom_save_hook=None
    ):
        """Register a recoverable with possible custom hooks.

        Arguments
        ---------
        name : str
            Unique name for recoverable. Used to map savefiles to objects.
        obj : instance
            The object to recover.
        custom_load_hook : callable
            Called to load the object's savefile. The function/method must be
            callable with signature (instance, path) using positional
            arguments. This is satisfied by for example: def load(self, path):
        custom_save_hook : callable
            Called to save the object's parameters. The function/method must
            be callable with signature (instance, path) using positional
            arguments. This is satisfied by for example: def saver(self, path):
        """
        self.recoverables[name] = obj
        if custom_load_hook is not None:
            self.custom_load_hooks[name] = custom_load_hook
        if custom_save_hook is not None:
            self.custom_save_hooks[name] = custom_save_hook

    def add_recoverables(self, recoverables):
        """Update the recoverables dict from the given mapping.

        Arguments
        ---------
        recoverables : mapping
            Objects to recover.
            They need a (unique) name: this is used to
            connect the parameters in a checkpoint to the correct
            recoverable. The name is also used in the filename of the
            savefile for the objects parameters.
        """
        if isinstance(recoverables, collections.abc.Mapping):
            self.recoverables.update(recoverables)
        else:
            rec = repr(recoverables)  # noqa: F841, rec is used in MSG
            MSG = f"Checkpointer needs a mapping (e.g. dict), \
                    got {rec} instead."
            raise AttributeError(MSG)

    def save_checkpoint(
        self, meta={}, end_of_epoch=True, name=None, verbosity=logging.INFO
    ):
        """Saves a checkpoint.

        The whole checkpoint becomes a directory.
        Saves each registered object's parameters in a separate file.
        Also a meta file is added. The meta file by default has just the
        unixtime (seconds since unix epoch), but you can add anything
        relevant yourself. The meta information is later used to pick the
        checkpoint to load.

        The value of end_of_epoch is saved in the meta. This can affect how
        epoch counters and dataset iterators load their state.

        Arguments
        ---------
        meta : mapping, optional
            A mapping which is added to the meta file in the checkpoint. The
            key "unixtime" is included by default.
        end_of_epoch : bool, optional
            Whether the checkpoint is at the end of an epoch. True by default.
            May affect loading.
        name : str, optional
            Specify a custom name for your checkpoint.
            The name will still have a prefix added. If no name is given,
            a name is created from a timestamp and a random unique id.
        verbosity : logging level
            Set logging level this save.

        Returns
        -------
        Checkpoint
            namedtuple [see above], the saved checkpoint.
        """
        if name is None:
            ckpt_dir = self._new_checkpoint_dirpath()
        else:
            ckpt_dir = self._custom_checkpoint_dirpath(name)
        os.makedirs(ckpt_dir)  # May raise FileExistsError, let it.
        saved_meta = self._save_checkpoint_metafile(
            ckpt_dir / METAFNAME, meta, end_of_epoch
        )
        saved_paramfiles = {}
        for name, obj in self.recoverables.items():
            objfname = f"{name}" + PARAMFILE_EXT
            savepath = ckpt_dir / objfname
            saved_paramfiles[name] = savepath
            # First see if object has custom load hook:
            if name in self.custom_save_hooks:
                self.custom_save_hooks[name](obj, savepath)
                continue
            # Otherwise find the default saver for that type:
            default_hook = get_default_hook(obj, DEFAULT_SAVE_HOOKS)
            if default_hook is not None:
                default_hook(obj, savepath)
                continue
            # If we got here, no custom hook or registered default hook
            MSG = f"Don't know how to save {type(obj)}. Register default hook \
                    or add custom hook for this object."
            raise RuntimeError(MSG)
        ckpt_type = "end-of-epoch" if end_of_epoch else "intra-epoch"
        logger.log(verbosity, f"Saved an {ckpt_type} checkpoint in {ckpt_dir}")
        return Checkpoint(ckpt_dir, saved_meta, saved_paramfiles)

    def save_and_keep_only(
        self,
        meta={},
        end_of_epoch=True,
        name=None,
        num_to_keep=1,
        keep_recent=True,
        importance_keys=[],
        max_keys=[],
        min_keys=[],
        ckpt_predicate=None,
        verbosity=logging.INFO,
    ):
        """Saves a checkpoint, then deletes the least important checkpoints.

        Essentially this combines ``save_checkpoint()`` and
        ``delete_checkpoints()`` in one call, providing short syntax.

        Arguments
        ---------
        meta : mapping, optional
            A mapping which is added to the meta file in the checkpoint. The
            key "unixtime" is included by default.
        end_of_epoch : bool, optional
            Whether the checkpoint is at the end of an epoch. True by default.
            May affect loading.
        name : str, optional
            Specify a custom name for your checkpoint.
            The name will still have a prefix added. If no name is given,
            a name is created from a timestamp and a random unique id.
        num_to_keep : int, optional
            Number of checkpoints to keep. Defaults to 1. This deletes all
            checkpoints remaining after filtering. Must be >=0.
        keep_recent : bool, optional
            Whether to keep the most recent ``num_to_keep`` checkpoints.
        importance_keys : list, optional
            A list of key functions used in sorting (see the sorted built-in).
            Each callable defines a sort order and num_to_keep checkpoints are
            kept for callable. The checkpoint with the highest keys are kept.
            The functions are passed Checkpoint namedtuples (see above).
        max_keys : list, optional
            A list of keys for which the *highest* value will be kept.
        min_keys : list, optional
            A list of keys for which the *lowest* value will be kept.
        ckpt_predicate : callable, optional
            Use this to exclude some checkpoints from deletion. Before any
            sorting, the list of checkpoints is filtered with this predicate.
            Only the checkpoints for which ckpt_predicate is True can be
            deleted. The function is called with Checkpoint namedtuples
            (see above).

        Returns
        -------
        None
            Unlike save_checkpoint, this does not return anything, since
            we cannot guarantee that the saved checkpoint actually survives
            deletion.
        """
        self.save_checkpoint(
            meta=meta, end_of_epoch=end_of_epoch, name=name, verbosity=verbosity
        )

        if keep_recent:
            importance_keys.append(ckpt_recency)
        self.delete_checkpoints(
            num_to_keep=num_to_keep,
            max_keys=max_keys,
            min_keys=min_keys,
            importance_keys=importance_keys,
            ckpt_predicate=ckpt_predicate,
            verbosity=verbosity,
        )

    def find_checkpoint(
        self,
        importance_key=None,
        max_key=None,
        min_key=None,
        ckpt_predicate=None,
    ):
        """Picks a particular checkpoint from all available checkpoints.

        If none of ``importance_key``, ``max_key``, and ``min_key`` is
        used, then most recent checkpoint will be returned. No more than
        one of them may be used.

        Most functionality is actually implemented in ``find_checkpoints()``
        but this is kept as a useful interface.

        Arguments
        ---------
        importance_key : callable, optional
            The key function used in sorting.
            The checkpoint with the highest returned value is picked.
            The function is called with Checkpoint namedtuples.
        max_key : str, optional
            The checkpoint with the highest value for this key will
            be returned. Only checkpoints with this key will be considered!
        min_key : str, optional
            The checkpoint with the lowest value for this key will
            be returned. Only checkpoints with this key will be considered!
        ckpt_predicate : callable, optional
            Before sorting, the list of
            checkpoints is filtered with this predicate.
            See the filter builtin.
            The function is called with Checkpoint namedtuples (see above).
            By default, all checkpoints are considered.

        Returns
        -------
        Checkpoint
            If found.
        None
            If no Checkpoints exist/remain after filtering.
        """
        ckpts_found = self.find_checkpoints(
            importance_key=importance_key,
            max_key=max_key,
            min_key=min_key,
            ckpt_predicate=ckpt_predicate,
            max_num_checkpoints=None,
        )
        if ckpts_found:
            return ckpts_found[0]
        else:
            return None

    def find_checkpoints(
        self,
        importance_key=None,
        max_key=None,
        min_key=None,
        ckpt_predicate=None,
        max_num_checkpoints=None,
    ):
        """Picks multiple checkpoints.

        If none of ``importance_key``, ``max_key``, and ``min_key`` is
        used, then the most recent checkpoints will be returned. No more than
        one of these may be used.

        Arguments
        ---------
        importance_key : callable, optional
            The key function used in sorting.
            The checkpoint with the highest returned value is picked.
            The function is called with Checkpoint namedtuples.
        max_key : str, optional
            The checkpoint with the highest value for this key will
            be returned. Only checkpoints with this key will be considered!
        min_key : str, optional
            The checkpoint with the lowest value for this key will
            be returned. Only checkpoints with this key will be considered!
        ckpt_predicate : callable, optional
            Before sorting, the list of
            checkpoints is filtered with this predicate.
            See the filter builtin.
            The function is called with Checkpoint namedtuples (see above).
            By default, all checkpoints are considered.
        max_num_checkpoints : int, None
            The maximum number of checkpoints to return, or None to return all
            found checkpoints.

        Returns
        -------
        list
            List containing at most the max specified number of Checkpoints.

        """
        if importance_key is None and min_key is None and max_key is None:
            importance_key = ckpt_recency

        if max_key and not importance_key:

            def importance_key(ckpt):
                "Defines the importance key."
                return ckpt.meta[max_key]

            def ckpt_predicate(ckpt, old_predicate=ckpt_predicate):
                "Checkpoints predicate."
                if old_predicate is not None:
                    return max_key in ckpt.meta and old_predicate(ckpt)
                else:
                    return max_key in ckpt.meta

        elif min_key and not importance_key:

            def importance_key(ckpt):
                "Defines the importance key."
                return -ckpt.meta[min_key]

            def ckpt_predicate(ckpt, old_predicate=ckpt_predicate):
                "Checkpoints predicate."
                if old_predicate is not None:
                    return min_key in ckpt.meta and old_predicate(ckpt)
                else:
                    return min_key in ckpt.meta

        elif min_key or max_key:
            raise ValueError(
                "Must specify only one of 'importance_key', 'max_key', "
                "and 'min_key'."
            )

        ckpts = self.list_checkpoints()
        ckpts = list(filter(ckpt_predicate, ckpts))
        # First sort by recency, so that importance being equal,
        # the most checkpoints are returned
        ckpts = sorted(ckpts, key=ckpt_recency, reverse=True)
        if ckpts:
            ranked_ckpts = sorted(ckpts, key=importance_key, reverse=True)
            # NOTE: apparently, you can also slice [:None],
            # and this is the same as [:], so the following if-else is not
            # strictly speaking needed. However, this feature does not seem to
            # be documented Python so I don't want to trust it.
            if max_num_checkpoints is not None:
                return ranked_ckpts[:max_num_checkpoints]
            else:  # No max number -> return all ckpts, but just sorted
                return ranked_ckpts
        else:
            return []  # Be explicit :)

    def recover_if_possible(
        self,
        importance_key=None,
        max_key=None,
        min_key=None,
        ckpt_predicate=None,
        device=None,
    ):
        """Picks a checkpoint and recovers from that, if one is found.

        If a checkpoint is not found, no recovery is run.

        If none of ``importance_key``, ``max_key``, and ``min_key`` is
        used, then most recent checkpoint will be returned. No more than
        one of them may be used.

        Arguments
        ---------
        importance_key : callable, optional
            The key function used in sorting.
            The checkpoint with the highest returned value is loaded.
            The function is called with Checkpoint namedtuples.
        max_key : str, optional
            The checkpoint with the highest value for this key will be loaded.
            Only checkpoints with this key will be considered!
        min_key : str, optional
            The checkpoint with the lowest value for this key will be loaded.
            Only checkpoints with this key will be considered!
        ckpt_predicate : callable, optional
            Before sorting, the list of
            checkpoints is filtered with this predicate.
            See the filter builtin.
            The function is called with Checkpoint namedtuples (see above).
            By default, all checkpoints are considered.
        device : torch.device
            Device to load models to.

        Returns
        -------
        Checkpoint
            If found.
        None
            If no Checkpoints exist/remain after filtering.
        """
        chosen_ckpt = self.find_checkpoint(
            importance_key, max_key, min_key, ckpt_predicate,
        )
        if chosen_ckpt is not None:
            self.load_checkpoint(chosen_ckpt, device)
        else:
            logger.info("Would load a checkpoint here, but none found yet.")
        return chosen_ckpt

    def load_checkpoint(self, checkpoint, device=None):
        """Loads the specified checkpoint.

        Arguments
        ---------
        checkpoint : Checkpoint
            Checkpoint to load.
        """
        self._call_load_hooks(checkpoint, device)

    def list_checkpoints(self):
        """List all checkpoints in the checkpoints directory.

        Returns
        -------
        list
            List of Checkpoint namedtuple (see above).
        """
        return self._construct_checkpoint_objects(self._list_checkpoint_dirs())

    # NOTE: * in arglist -> keyword only arguments
    def delete_checkpoints(
        self,
        *,
        num_to_keep=1,
        min_keys=None,
        max_keys=None,
        importance_keys=[ckpt_recency],
        ckpt_predicate=None,
        verbosity=logging.INFO,
    ):
        """Deletes least important checkpoints.

        Since there can be many ways to define importance (e.g. lowest WER,
        lowest loss), the user should provide a list of sort key functions,
        each defining a particular importance order. In essence, each
        importance key function extracts one importance metric (higher is more
        important). For each of these orders, num_to_keep checkpoints are kept.
        However if there is overlap between each orders' preserved checkpoints,
        the additional checkpoints are not preserved, so the total number of
        preserved checkpoints can be less than::

            num_to_keep * len(importance_keys)

        Arguments
        ---------
        num_to_keep : int, optional
            Number of checkpoints to keep.
            Defaults to 10. You choose to keep 0. This deletes all
            checkpoints remaining after filtering. Must be >=0
        min_keys : list, optional
            List of strings representing keys in the meta. The lowest of
            these values will be kept, up to num_to_keep.
        max_keys : list, optional
            List of strings representing keys in the meta. The highest of
            these values will be kept, up to num_to_keep.
        importance_keys : list, optional
            A list of key functions used in sorting (see the sorted built-in).
            Each callable defines a sort order and num_to_keep checkpoints are
            kept for  callable. To be clear, those with the highest key are
            kept.
            The functions are called with Checkpoint namedtuples
            (see above). See also the default (ckpt_recency,
            above). The default deletes all but the latest checkpoint.
        ckpt_predicate : callable, optional
            Use this to exclude some checkpoints from deletion. Before any
            sorting, the list of checkpoints is filtered with this predicate.
            Only the checkpoints for which ckpt_predicate is True can be
            deleted. The function is called with Checkpoint namedtuples
            (see above).
        verbosity : logging level
            Set logging level for this deletion.

        Note
        ----
        Must be called with keyword arguments, as a signoff that you
        know what you are doing. Deletion is permanent.
        """
        if num_to_keep < 0:
            raise ValueError("Number of checkpoints to keep must be positive.")

        # Build a list of potential deletions and protected checkpoints
        potential_deletions = set()
        protected_checkpoints = set()
        keys = [{"min_key": key} for key in min_keys or []]
        keys.extend([{"max_key": key} for key in max_keys or []])
        keys.extend([{"importance_key": key} for key in importance_keys])

        # Don't consider checkpoints for deletion that don't have a listed key
        for key_kwargs in keys:
            key_kwargs["ckpt_predicate"] = ckpt_predicate
            potential_deletions.update(self.find_checkpoints(**key_kwargs))
            protected_checkpoints.update(
                self.find_checkpoints(
                    max_num_checkpoints=num_to_keep, **key_kwargs
                )
            )

        # Delete unprotected checkpoints
        for ckpt in potential_deletions:
            if ckpt not in protected_checkpoints:
                Checkpointer._delete_checkpoint(ckpt, verbosity=verbosity)

    @staticmethod
    def _delete_checkpoint(checkpoint, verbosity=logging.INFO):
        if not Checkpointer._is_checkpoint_dir(checkpoint.path):
            raise RuntimeError("Checkpoint does not appear valid for deletion.")
        shutil.rmtree(checkpoint.path)
        logger.log(verbosity, f"Deleted checkpoint in {checkpoint.path}")

    def _call_load_hooks(self, checkpoint, device=None):
        # This internal function finds the correct hook to call for every
        # recoverable, and calls it.
        logger.info(f"Loading a checkpoint from {checkpoint.path}")
        end_of_epoch = checkpoint.meta["end-of-epoch"]
        for name, obj in self.recoverables.items():
            # NOTE: We want the checkpoint namedtuple to have the paramfile
            # paths for each recoverable.
            # In some rare case, the user can e.g. add a path there manually.
            try:
                loadpath = checkpoint.paramfiles[name]
            except KeyError:
                if self.allow_partial_load:
                    continue
                elif "dataloader" in name:
                    MSG = f"Loading checkpoint from {checkpoint.path}, \
                            but missing a load path for {name}"
                    warnings.warn(MSG, UserWarning)
                    continue
                else:
                    MSG = f"Loading checkpoint from {checkpoint.path}, \
                            but missing a load path for {name}"
                    raise RuntimeError(MSG)

            # First see if object has custom load hook:
            if name in self.custom_load_hooks:
                self.custom_load_hooks[name](
                    obj, loadpath, end_of_epoch, device
                )
                continue
            # Otherwise find the default saver for that type:
            default_hook = get_default_hook(obj, DEFAULT_LOAD_HOOKS)
            if default_hook is not None:
                default_hook(obj, loadpath, end_of_epoch, device)
                continue
            # If we got here, no custom hook or registered default hook exists
            MSG = f"Don't know how to load {type(obj)}. Register default hook \
                    or add custom hook for this object."
            raise RuntimeError(MSG)

    def _list_checkpoint_dirs(self):
        # This internal method returns a list of individual checkpoint
        # directory paths in the top checkpoint directory
        return [
            x
            for x in self.checkpoints_dir.iterdir()
            if Checkpointer._is_checkpoint_dir(x)
        ]

    @staticmethod
    def _construct_checkpoint_objects(checkpoint_dirs):
        # This internal method takes a list of individual checkpoint
        # directory paths (as produced by _list_checkpoint_dirs)
        checkpoints = []
        for ckpt_dir in checkpoint_dirs:
            with open(ckpt_dir / METAFNAME) as fi:
                meta = yaml.load(fi, Loader=yaml.Loader)
            paramfiles = {}
            for ckptfile in ckpt_dir.iterdir():
                if ckptfile.suffix == PARAMFILE_EXT:
                    paramfiles[ckptfile.stem] = ckptfile
            checkpoints.append(Checkpoint(ckpt_dir, meta, paramfiles))
        return checkpoints

    @staticmethod
    def _is_checkpoint_dir(path):
        # This internal method verifies whether a given path points to a
        # directory that holds a checkpoint.
        path = pathlib.Path(path)
        if not path.is_dir():
            return False
        if not path.name.startswith(CKPT_PREFIX):
            return False
        return (path / METAFNAME).exists()

    def _new_checkpoint_dirpath(self):
        # This internal method creates a checkpoint name and returns a path
        # to that directory (but does not create the directory!)
        t = time.time()
        stamp = time.strftime("%Y-%m-%d+%H-%M-%S", time.localtime(t))
        suffix_num = 0
        while (
            self.checkpoints_dir / f"{CKPT_PREFIX}+{stamp}+{suffix_num:02d}"
        ).exists():
            suffix_num += 1
        return self.checkpoints_dir / f"{CKPT_PREFIX}+{stamp}+{suffix_num:02d}"

    def _custom_checkpoint_dirpath(self, name):
        # This internal method creates a checkpoint name based on a given
        # custom name and returns a path to that directory (but does not
        # create the directory!)
        return self.checkpoints_dir / f"{CKPT_PREFIX}+{name}"

    def _save_checkpoint_metafile(
        self, fpath, meta_to_include={}, end_of_epoch=True
    ):
        # This internal method saves the meta information in the given path
        meta = {"unixtime": time.time(), "end-of-epoch": end_of_epoch}
        meta.update(meta_to_include)
        with open(fpath, "w") as fo:
            fo.write("# yamllint disable\n")
            fo.write(yaml.dump(meta))
        return meta


def average_state_dicts(state_dicts):
    """Produces an average state_dict from an iterator over state_dicts.

    Note that at one time, this keeps two of the state_dicts in memory, which
    is the minimum memory requirement.

    Arguments
    ---------
    state_dicts : iterator, list
        The state_dicts to average.

    Returns
    -------
    state_dict
        The averaged state_dict.
    """
    iterator = iter(state_dicts)
    try:
        running_sum = next(iterator)
    except StopIteration:
        raise ValueError("No state dicts to average.")
    num_dicts = 1
    with torch.no_grad():
        # First sum all state_dicts together:
        for state_dict in iterator:
            for pname, param in state_dict.items():
                running_sum[pname] += param.data
            num_dicts += 1
        # Finally, divide by number of dicts:
        for pname, param in running_sum.items():
            running_sum[pname] = param.data / float(num_dicts)
    return running_sum



#!/usr/bin/python
# -*- coding: utf-8 -*-
from .adversarial_adapter import AdversarialAdapter
import torch.nn as nn
import torch
import math
from .grl import GradientReverseLayer
import logging

class RandomLayer(nn.Module):
    ''' A non-trainable random layer, see paper: Conditional Adversarial Domain Adaptation

    '''
    def __init__(self, input_dim_list, output_dim):
        ''' init method

        :param input_dim_list: input dim list
        :param output_dim: output dim
        '''
        super(RandomLayer, self).__init__()
        self.input_dim_list = input_dim_list
        self.input_num = len(input_dim_list)
        self.output_dim = output_dim
        self.random_matrix = [torch.randn(input_dim_list[i], output_dim)  # random matrix (no trainable)
                              for i in range(self.input_num)]

    def forward(self, input_list):
        ''' forward

        :param input_list: input tensor list, must match 'input_dim_list'
        :return:
        '''
        return_list = [torch.mm(input_list[i], self.random_matrix[i]) for i in range(self.input_num)]
        return_tensor = return_list[0] / math.pow(float(self.output_dim), 1.0 / len(return_list))
        for single in return_list[1:]:
            return_tensor = torch.mul(return_tensor, single)
        return return_tensor

    def cuda(self):
        ''' move random matrix to cuda

        :return:
        '''
        super(RandomLayer, self).cuda()
        self.random_matrix = [val.cuda() for val in self.random_matrix]

class CDANAdapter(AdversarialAdapter):
    ''' CDAN , see paper: Conditional Adversarial Domain Adaptation

    '''
    def __init__(self,in_feature, hidden_size,dropout_rate,grl_coeff_alpha,grl_coeff_high,max_iter,
                 backbone_output_size,enable_random_layer,enable_entropy_weight):
        ''' Init method

        :param in_feature: adversarial network input
        :param hidden_size: hidden size
        :param dropout_rate: dropout rate
        :param grl_coeff_alpha: GradientReverseLayer alpha argument to calculate coef
        :param grl_coeff_high: GradientReverseLayer coef high
        :param max_iter: max iter for one epoch
        :param backbone_output_size: backbone output size
        :param enable_random_layer: whether using random layer
        :param enable_entropy_weight: whether using entropy as weight
        '''
        super(CDANAdapter, self).__init__(in_feature, hidden_size,dropout_rate,grl_coeff_alpha,grl_coeff_high,max_iter)
        if enable_random_layer:
            self._random_layer = RandomLayer([in_feature, backbone_output_size], in_feature)
            logging.info("CDAN enable random layer")
        else:
            self._random_layer = None
        self.entropy_grl = GradientReverseLayer(coeff_alpha=grl_coeff_alpha, coeff_high=grl_coeff_high,
                                        max_iter=max_iter, enable_step=self.training)

        self._enable_entropy_weight = enable_entropy_weight
        if self._enable_entropy_weight:
            logging.info("CDAN enable entropy weight")

    def _forward_input(self,adapter_input,backbone_output):
        ''' get forward input

        :param adapter_input: original adapter input
        :param backbone_output: backbone output
        :return: modified adapter input
        '''
        if self._random_layer is None:
            size = backbone_output.size(1) * adapter_input.size(1)
            return torch.bmm(backbone_output.unsqueeze(2), adapter_input.unsqueeze(1)).view(-1, size)  # outer product of feature and output
        else: # random layer for computation efficiency
            return self._random_layer.forward([adapter_input,backbone_output])          # fixed random layer

    def forward(self, x,**kwargs):
        ''' CDAN forward

        :param args: args
        :param kwargs: kwargs, must contain backbone_output
        :return:
        '''
        adapter_input = x
        backbone_output = kwargs['backbone_output']
        new_input = self._forward_input(adapter_input,backbone_output)
        return super(CDANAdapter, self).forward(new_input)

    def _normalized_entropy_weight(self,backbone_output):
        ''' get the normalized entropy as weight

        :param backbone_output: backbone raw out, which is not the probability
        :return: normalized weight
        '''
        backbone_prediction = nn.Softmax(dim=1)(backbone_output)
        entropy = torch.mean(torch.special.entr(backbone_prediction), dim=1)
        entropy = self.entropy_grl(entropy)
        entropy = 1.0 + torch.exp(-entropy)  # entropy as weight
        ########### weight normalization ###########
        sum_weight = torch.sum(entropy).detach().item()  # simplify gradient
        return entropy/sum_weight

    def loss(self,output_prob,label,**kwargs):
        ''' adapter loss function

        :param output_prob: output probability
        :param label: ground truth
        :param kwargs: kwargs, must contain backbone_output
        :return: loss
        '''
        if self._enable_entropy_weight:  # quantify the uncertainty of classifier predictions by the entropy, to emphasize those easy-to-transfer exmples
            backbone_output = kwargs['backbone_output']
            normalized_weight = self._normalized_entropy_weight(backbone_output)
            return torch.sum(normalized_weight * nn.BCELoss(reduction='none')(output_prob, label))
        else:
            return nn.BCELoss()(output_prob, label)


# coding=utf-8
# Copyright (c) 2022, Intel Corporation

# MIT License
# Copyright (c) Microsoft Corporation.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

# =======================================================================================
# MIT license
# =======================================================================================
# - [Swin Transformer](https://github.com/microsoft/swin-transformer)
# - [CLIP](https://github.com/openai/CLIP)

# =======================================================================================
# Apache license 2.0
# =======================================================================================
# - [LeViT](https://github.com/facebookresearch/LeViT)
# - [pytorch-image-models](https://github.com/rwightman/pytorch-image-models)

# =======================================================================================
# BSD-style license
# =======================================================================================
# - [PyTorch](https://github.com/pytorch/pytorch)

import numpy as np
from numpy.random import Generator, PCG64

RNG = None


class AugRandomContext:
    def __init__(self, seed):
        self.seed = seed

    def __enter__(self):
        global RNG
        assert RNG is None
        RNG = Generator(PCG64(seed=self.seed))

    def __exit__(self, *_):
        global RNG
        RNG = None


class random:
    # inline: random module
    @staticmethod
    def random():
        return RNG.random()

    @staticmethod
    def uniform(a, b):
        return random.random() * (b - a) + a

    @staticmethod
    def randint(a, b):
        # [low, high]
        return min(int(random.random() * (b - a + 1)) + a, b)

    @staticmethod
    def gauss(mu, sigma):
        return RNG.normal(mu, sigma)


class np_random:
    # numpy.random
    @staticmethod
    def choice(a, size, *args, **kwargs):
        return RNG.choice(a, size, *args, **kwargs)

    @staticmethod
    def randint(low, high, size=None, dtype=int):
        # [low, high)
        if size is None:
            return dtype(random.randint(low, high - 1))
        out = [random.randint(low, high - 1) for _ in range(size)]
        return np.array(out, dtype=dtype)

    @staticmethod
    def rand(*shape):
        return RNG.random(shape)

    @staticmethod
    def beta(a, b, size=None):
        return RNG.beta(a, b, size=size)


if __name__ == '__main__':
    for _ in range(2):
        with AugRandomContext(seed=0):
            print(np_random.randint(-100, 100, size=10))
        with AugRandomContext(seed=1):
            print(np_random.randint(-100, 100, size=10))



from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import math
import logging

import torch
from torch import nn

from e2eAIOK.DeNas.module.nlp.Linear_super import LinearSuper as SuperLinear


logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s',
                    datefmt='%m/%d/%Y %H:%M:%S',
                    level=logging.INFO)
logger = logging.getLogger(__name__)

def gelu(x):
    """Implementation of the gelu activation function.
        For information: OpenAI GPT's gelu is slightly different (and gives slightly different results):
        0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
        Also see https://arxiv.org/abs/1606.08415
    """
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))

def swish(x):
    return x * torch.sigmoid(x)

ACT2FN = {"gelu": gelu, "relu": torch.nn.functional.relu, "swish": swish}

class SuperBertIntermediate(nn.Module):
    def __init__(self, config):
        super(SuperBertIntermediate, self).__init__()
        self.dense = SuperLinear(config.hidden_size, config.intermediate_size)
        if isinstance(config.hidden_act, str) or (sys.version_info[0] == 2 and isinstance(config.hidden_act, unicode)):
            self.intermediate_act_fn = ACT2FN[config.hidden_act]
        else:
            self.intermediate_act_fn = config.hidden_act

    def set_sample_config(self, sample_embed_dim, intermediate_size):
        self.dense.set_sample_config(sample_embed_dim, intermediate_size)

    def calc_sampled_param_num(self):
        dense_numel = self.dense.calc_sampled_param_num()

        #logger.info('dense_numel: {}\n'.format(dense_numel))
        return dense_numel

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states



__all__ = ['config','data','decoders','lib','scripts','trainer','utils','asr_model_builder','asr_trainer','init_asr_parser','supernet_asr','train','TransformerBase','TransformerLM', 'model_builder_denas_asr']


#!/usr/bin/python
# -*- coding: utf-8 -*-

import torch.nn as nn
from torch.autograd import Function
import torch
from typing import Optional, Any, Tuple
import numpy as np

class GradientReverseOp(Function):
    ''' Gradient Reverse Operation

    '''
    @staticmethod
    def forward(ctx: Any, input: torch.Tensor, coeff: Optional[float] = 1.) -> torch.Tensor:
        ''' forward function

        :param ctx: context
        :param input: input tensor
        :param coeff: coeff to warm start Gradient Reverse Operation
        :return: output
        '''
        ctx.coeff = coeff # save coeff for backward using
        return input

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> Tuple[torch.Tensor, Any]:
        return grad_output.neg() * ctx.coeff, None

class GradientReverseLayer(nn.Module):
    ''' Gradient Reverse Layer

    '''
    def __init__(self, coeff_alpha: float,coeff_high: float,max_iter: int, enable_step: bool):
        ''' Init method.

        :param coeff_alpha: control warm curve shape. The larger, the steeper
        :param coeff_high: control high of the curve shape
        :param max_iter: max iter for one epoch
        :param enable_step: frozen iter_num or not
        '''
        super(GradientReverseLayer, self).__init__()
        self.coeff_alpha = coeff_alpha
        self.coeff_high = coeff_high
        self.iter_num = 0
        self.max_iter = max_iter
        self.enable_step = enable_step

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """"""
        self.coeff = 2.0*self.coeff_high /(1.0 + np.exp(-self.coeff_alpha * self.iter_num / self.max_iter))- self.coeff_high

        if self.enable_step: # increase iter_num
            self.iter_num += 1
        return GradientReverseOp.apply(input, self.coeff)




import torch
from torch import nn
from torch.nn import Parameter
import torch.nn.functional as F

from e2eAIOK.DeNas.module.attention_base import AttentionBase
from e2eAIOK.DeNas.module.cv.Linear_super import LinearSuper
from e2eAIOK.DeNas.module.cv.qkv_super import qkv_super
from e2eAIOK.DeNas.module.cv.utils import trunc_normal_


def softmax(x, dim, onnx_trace=False):
    if onnx_trace:
        return F.softmax(x.float(), dim=dim)
    else:
        return F.softmax(x, dim=dim, dtype=torch.float32)

class RelativePosition2D_super(nn.Module):

    def __init__(self, num_units, max_relative_position):
        super().__init__()

        self.num_units = num_units
        self.max_relative_position = max_relative_position
        # The first element in embeddings_table_v is the vertical embedding for the class
        self.embeddings_table_v = nn.Parameter(torch.randn(max_relative_position * 2 + 2, num_units))
        self.embeddings_table_h = nn.Parameter(torch.randn(max_relative_position * 2 + 2, num_units))

        trunc_normal_(self.embeddings_table_v, std=.02)
        trunc_normal_(self.embeddings_table_h, std=.02)

        self.sample_head_dim = None
        self.sample_embeddings_table_h = None
        self.sample_embeddings_table_v = None

    def set_sample_config(self, sample_head_dim):
        self.sample_head_dim = sample_head_dim
        self.sample_embeddings_table_h = self.embeddings_table_h[:,:sample_head_dim]
        self.sample_embeddings_table_v = self.embeddings_table_v[:,:sample_head_dim]

    def calc_sampled_param_num(self):
        return self.sample_embeddings_table_h.numel() + self.sample_embeddings_table_v.numel()

    def forward(self, length_q, length_k):
        # remove the first cls token distance computation
        length_q = length_q - 1
        length_k = length_k - 1
        range_vec_q = torch.arange(length_q)
        range_vec_k = torch.arange(length_k)
        # compute the row and column distance
        distance_mat_v = (range_vec_k[None, :] // int(length_q ** 0.5 )  - range_vec_q[:, None] // int(length_q ** 0.5 ))
        distance_mat_h = (range_vec_k[None, :] % int(length_q ** 0.5 ) - range_vec_q[:, None] % int(length_q ** 0.5 ))
        # clip the distance to the range of [-max_relative_position, max_relative_position]
        distance_mat_clipped_v = torch.clamp(distance_mat_v, -self.max_relative_position, self.max_relative_position)
        distance_mat_clipped_h = torch.clamp(distance_mat_h, -self.max_relative_position, self.max_relative_position)

        # translate the distance from [1, 2 * max_relative_position + 1], 0 is for the cls token
        final_mat_v = distance_mat_clipped_v + self.max_relative_position + 1
        final_mat_h = distance_mat_clipped_h + self.max_relative_position + 1
        # pad the 0 which represent the cls token
        final_mat_v = torch.nn.functional.pad(final_mat_v, (1,0,1,0), "constant", 0)
        final_mat_h = torch.nn.functional.pad(final_mat_h, (1,0,1,0), "constant", 0)

        final_mat_v = torch.LongTensor(final_mat_v)
        final_mat_h = torch.LongTensor(final_mat_h)
        # get the embeddings with the corresponding distance
        embeddings = self.sample_embeddings_table_v[final_mat_v] + self.sample_embeddings_table_h[final_mat_h]

        return embeddings

class AttentionSuper(AttentionBase):
    def __init__(self, super_embed_dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0., normalization = False, relative_position = False,
                 num_patches = None, max_relative_position=14, scale=False, change_qkv = False):
        super().__init__()
        self.num_heads = num_heads
        head_dim = super_embed_dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5
        self.super_embed_dim = super_embed_dim

        self.fc_scale = scale
        self.change_qkv = change_qkv
        if change_qkv:
            self.qkv = qkv_super(super_embed_dim, 3 * super_embed_dim, bias=qkv_bias)
        else:
            self.qkv = LinearSuper(super_embed_dim, 3 * super_embed_dim, bias=qkv_bias)

        self.relative_position = relative_position
        if self.relative_position:
            self.rel_pos_embed_k = RelativePosition2D_super(super_embed_dim //num_heads, max_relative_position)
            self.rel_pos_embed_v = RelativePosition2D_super(super_embed_dim //num_heads, max_relative_position)
        self.max_relative_position = max_relative_position
        self.sample_qk_embed_dim = None
        self.sample_v_embed_dim = None
        self.sample_num_heads = None
        self.sample_scale = None
        self.sample_in_embed_dim = None

        self.proj = LinearSuper(super_embed_dim, super_embed_dim)

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj_drop = nn.Dropout(proj_drop)

    def set_sample_config(self, sample_q_embed_dim=None, sample_num_heads=None, sample_in_embed_dim=None):

        self.sample_in_embed_dim = sample_in_embed_dim
        self.sample_num_heads = sample_num_heads
        if not self.change_qkv:
            self.sample_qk_embed_dim = self.super_embed_dim
            self.sample_scale = (sample_in_embed_dim // self.sample_num_heads) ** -0.5

        else:
            self.sample_qk_embed_dim = sample_q_embed_dim
            self.sample_scale = (self.sample_qk_embed_dim // self.sample_num_heads) ** -0.5

        self.qkv.set_sample_config(sample_in_dim=sample_in_embed_dim, sample_out_dim=3*self.sample_qk_embed_dim)
        self.proj.set_sample_config(sample_in_dim=self.sample_qk_embed_dim, sample_out_dim=sample_in_embed_dim)
        if self.relative_position:
            self.rel_pos_embed_k.set_sample_config(self.sample_qk_embed_dim // sample_num_heads)
            self.rel_pos_embed_v.set_sample_config(self.sample_qk_embed_dim // sample_num_heads)
    def calc_sampled_param_num(self):

        return 0
    def get_complexity(self, sequence_length):
        total_flops = 0
        total_flops += self.qkv.get_complexity(sequence_length)
        # attn
        total_flops += sequence_length * sequence_length * self.sample_qk_embed_dim
        # x
        total_flops += sequence_length * sequence_length * self.sample_qk_embed_dim
        total_flops += self.proj.get_complexity(sequence_length)
        if self.relative_position:
            total_flops += self.max_relative_position * sequence_length * sequence_length + sequence_length * sequence_length / 2.0
            total_flops += self.max_relative_position * sequence_length * sequence_length + sequence_length * self.sample_qk_embed_dim / 2.0
        return total_flops

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.sample_num_heads, -1).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]   # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.sample_scale
        if self.relative_position:
            r_p_k = self.rel_pos_embed_k(N, N)
            attn = attn + (q.permute(2, 0, 1, 3).reshape(N, self.sample_num_heads * B, -1) @ r_p_k.transpose(2, 1)) \
                .transpose(1, 0).reshape(B, self.sample_num_heads, N, N) * self.sample_scale

        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1,2).reshape(B, N, -1)
        if self.relative_position:
            r_p_v = self.rel_pos_embed_v(N, N)
            attn_1 = attn.permute(2, 0, 1, 3).reshape(N, B * self.sample_num_heads, -1)
            # The size of attention is (B, num_heads, N, N), reshape it to (N, B*num_heads, N) and do batch matmul with
            # the relative position embedding of V (N, N, head_dim) get shape like (N, B*num_heads, head_dim). We reshape it to the
            # same size as x (B, num_heads, N, hidden_dim)
            x = x + (attn_1 @ r_p_v).transpose(1, 0).reshape(B, self.sample_num_heads, N, -1).transpose(2,1).reshape(B, N, -1)

        if self.fc_scale:
            x = x * (self.super_embed_dim / self.sample_qk_embed_dim)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x



from e2eAIOK.DeNas.utils import decode_arch_tuple
from e2eAIOK.DeNas.cv.third_party.ZenNet import DeMainNet
from e2eAIOK.common.trainer.model.model_builder_cv import ModelBuilderCV
from e2eAIOK.DeNas.cv.supernet_transformer import Vision_TransformerSuper

class ModelBuilderCVDeNas(ModelBuilderCV):
    def __init__(self, cfg):
        super().__init__(cfg)
    
    def _init_model(self):
        if self.cfg.best_model_structure != None:
            with open(self.cfg.best_model_structure, 'r') as f:
                arch = f.readlines()[-1]
        else:
            raise RuntimeError(f"model structure string not found")
        
        if self.cfg.domain == 'cnn':
            model = DeMainNet(num_classes=self.cfg.num_classes, plainnet_struct=arch, no_create=False)
        elif self.cfg.domain == 'vit':
            model = Vision_TransformerSuper(img_size=self.cfg.input_size,
                                patch_size=self.cfg.patch_size,
                                embed_dim=self.cfg['SUPERNET']['EMBED_DIM'], depth=self.cfg['SUPERNET']['DEPTH'],
                                num_heads=self.cfg['SUPERNET']['NUM_HEADS'],mlp_ratio=self.cfg['SUPERNET']['MLP_RATIO'],
                                qkv_bias=True, drop_rate=self.cfg.drop,
                                drop_path_rate=self.cfg.drop_path,
                                gp=self.cfg.gp,
                                num_classes=self.cfg.num_classes,
                                max_relative_position=self.cfg.max_relative_position,
                                relative_position=self.cfg.relative_position,
                                change_qkv=self.cfg.change_qkv, abs_pos=not self.cfg.no_abs_pos)
            depth, mlp_ratio, num_heads, embed_dim = decode_arch_tuple(arch)
            model_config = {}
            model_config['layer_num'] = depth
            model_config['mlp_ratio'] = mlp_ratio
            model_config['num_heads'] = num_heads
            model_config['embed_dim'] = [embed_dim]*depth
            n_parameters = model.get_sampled_params_numel(model_config)
            print("model parameters size: {}".format(n_parameters))
        return model
        


import torch
from e2eAIOK.DeNas.asr.lib.CNN import Conv2d
from e2eAIOK.DeNas.asr.lib.containers import Sequential
from e2eAIOK.DeNas.module.asr.normalization import LayerNorm


class ConvolutionFrontEnd(Sequential):
    """This is a module to ensemble a convolution (depthwise) encoder with or
    without residual connection.

     Arguments
    ----------
    out_channels: int
        Number of output channels of this model (default 640).
    out_channels: Optional(list[int])
        Number of output channels for each of block.
    kernel_size: int
        Kernel size of convolution layers (default 3).
    strides: Optional(list[int])
        Striding factor for each block, this stride is applied at the last convolution layer at each block.
    num_blocks: int
        Number of block (default 21).
    num_per_layers: int
        Number of convolution layers for each block (default 5).
    dropout: float
        Dropout (default 0.15).
    activation: torch class
        Activation function for each block (default Swish).
    norm: torch class
        Normalization to regularize the model (default BatchNorm1d).
    residuals: Optional(list[bool])
        Whether apply residual connection at each block (default None).
    """

    def __init__(
        self,
        input_shape,
        num_blocks=3,
        num_layers_per_block=5,
        out_channels=[128, 256, 512],
        kernel_sizes=[3, 3, 3],
        strides=[1, 2, 2],
        dilations=[1, 1, 1],
        residuals=[True, True, True],
        conv_module=Conv2d,
        activation=torch.nn.LeakyReLU,
        norm=LayerNorm,
        dropout=0.1,
    ):
        super().__init__(input_shape=input_shape)
        for i in range(num_blocks):
            self.append(
                ConvBlock,
                num_layers=num_layers_per_block,
                out_channels=out_channels[i],
                kernel_size=kernel_sizes[i],
                stride=strides[i],
                dilation=dilations[i],
                residual=residuals[i],
                conv_module=conv_module,
                activation=activation,
                norm=norm,
                dropout=dropout,
                layer_name=f"convblock_{i}",
            )


class ConvBlock(torch.nn.Module):
    """An implementation of convolution block with 1d or 2d convolutions (depthwise).

    Arguments
    ----------
    out_channels : int
        Number of output channels of this model (default 640).
    kernel_size : int
        Kernel size of convolution layers (default 3).
    strides : int
        Striding factor for this block (default 1).
    num_layers : int
        Number of depthwise convolution layers for this block.
    activation : torch class
        Activation function for this block.
    norm : torch class
        Normalization to regularize the model (default BatchNorm1d).
    residuals: bool
        Whether apply residual connection at this block (default None).

    Example
    -------
    >>> x = torch.rand((8, 30, 10))
    >>> conv = ConvBlock(2, 16, input_shape=x.shape)
    >>> out = conv(x)
    >>> x.shape
    torch.Size([8, 30, 10])
    """

    def __init__(
        self,
        num_layers,
        out_channels,
        input_shape,
        kernel_size=3,
        stride=1,
        dilation=1,
        residual=False,
        conv_module=Conv2d,
        activation=torch.nn.LeakyReLU,
        norm=None,
        dropout=0.1,
    ):
        super().__init__()

        self.convs = Sequential(input_shape=input_shape)

        for i in range(num_layers):
            self.convs.append(
                conv_module,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride if i == num_layers - 1 else 1,
                dilation=dilation,
                layer_name=f"conv_{i}",
            )
            if norm is not None:
                self.convs.append(norm, layer_name=f"norm_{i}")
            self.convs.append(activation(), layer_name=f"act_{i}")
            self.convs.append(
                torch.nn.Dropout(dropout), layer_name=f"dropout_{i}"
            )

        self.reduce_conv = None
        self.drop = None
        if residual:
            self.reduce_conv = Sequential(input_shape=input_shape)
            self.reduce_conv.append(
                conv_module,
                out_channels=out_channels,
                kernel_size=1,
                stride=stride,
                layer_name="conv",
            )
            self.reduce_conv.append(norm, layer_name="norm")
            self.drop = torch.nn.Dropout(dropout)

    def forward(self, x):
        """ Processes the input tensor x and returns an output tensor."""
        out = self.convs(x)
        if self.reduce_conv:
            out = out + self.reduce_conv(x)
            out = self.drop(out)

        return out



#!/usr/bin/python
# -*- coding: utf-8 -*-
from .adversarial.cdan_adapter import CDANAdapter
from .adversarial.dann_adapter import DANNAdapter
from .adversarial.DA_Loss import CACDomainAdversarialLoss
import logging

def createAdapter(adapter_name, **kwargs):
    ''' create adapter by name

    :param adapter_name: adapter name
    :param kwargs: kwargs to create adapter
    :return: a adapter model
    '''
    if adapter_name == 'DANN':
        return DANNAdapter(**kwargs)
    elif adapter_name == 'CDAN':
        return CDANAdapter(**kwargs)
    elif adapter_name == "CAC_UNet":
        return CACDomainAdversarialLoss(**kwargs)
    else:
        logging.error("[%s] is not supported" % adapter_name)
        raise NotImplementedError("[%s] is not supported" % adapter_name)


import os
import re
import shutil
import urllib.request
import collections.abc
import torch
import tqdm
import pathlib

from e2eAIOK.DeNas.asr.utils.utils import is_main_process, ddp_barrier


def undo_padding(batch, lengths):
    """Produces Python lists given a batch of sentences with
    their corresponding relative lengths.

    Arguments
    ---------
    batch : tensor
        Batch of sentences gathered in a batch.
    lengths : tensor
        Relative length of each sentence in the batch.
    """
    batch_max_len = batch.shape[1]
    as_list = []
    for seq, seq_length in zip(batch, lengths):
        actual_size = int(torch.round(seq_length * batch_max_len))
        seq_true = seq.narrow(0, 0, actual_size)
        as_list.append(seq_true.tolist())
    return as_list


def get_all_files(
    dirName, match_and=None, match_or=None, exclude_and=None, exclude_or=None
):
    """Returns a list of files found within a folder.

    Different options can be used to restrict the search to some specific
    patterns.

    Arguments
    ---------
    dirName : str
        The directory to search.
    match_and : list
        A list that contains patterns to match. The file is
        returned if it matches all the entries in `match_and`.
    match_or : list
        A list that contains patterns to match. The file is
        returned if it matches one or more of the entries in `match_or`.
    exclude_and : list
        A list that contains patterns to match. The file is
        returned if it matches none of the entries in `exclude_and`.
    exclude_or : list
        A list that contains pattern to match. The file is
        returned if it fails to match one of the entries in `exclude_or`.
    """

    # Match/exclude variable initialization
    match_and_entry = True
    match_or_entry = True
    exclude_or_entry = False
    exclude_and_entry = False

    # Create a list of file and sub directories
    listOfFile = os.listdir(dirName)
    allFiles = list()

    # Iterate over all the entries
    for entry in listOfFile:

        # Create full path
        fullPath = os.path.join(dirName, entry)

        # If entry is a directory then get the list of files in this directory
        if os.path.isdir(fullPath):
            allFiles = allFiles + get_all_files(
                fullPath,
                match_and=match_and,
                match_or=match_or,
                exclude_and=exclude_and,
                exclude_or=exclude_or,
            )
        else:

            # Check match_and case
            if match_and is not None:
                match_and_entry = False
                match_found = 0

                for ele in match_and:
                    if ele in fullPath:
                        match_found = match_found + 1
                if match_found == len(match_and):
                    match_and_entry = True

            # Check match_or case
            if match_or is not None:
                match_or_entry = False
                for ele in match_or:
                    if ele in fullPath:
                        match_or_entry = True
                        break

            # Check exclude_and case
            if exclude_and is not None:
                match_found = 0

                for ele in exclude_and:
                    if ele in fullPath:
                        match_found = match_found + 1
                if match_found == len(exclude_and):
                    exclude_and_entry = True

            # Check exclude_or case
            if exclude_or is not None:
                exclude_or_entry = False
                for ele in exclude_or:
                    if ele in fullPath:
                        exclude_or_entry = True
                        break

            # If needed, append the current file to the output list
            if (
                match_and_entry
                and match_or_entry
                and not (exclude_and_entry)
                and not (exclude_or_entry)
            ):
                allFiles.append(fullPath)

    return allFiles


def download_file(
    source, dest, unpack=False, dest_unpack=None, replace_existing=False
):
    """Downloads the file from the given source and saves it in the given
    destination path.

     Arguments
    ---------
    source : path or url
        Path of the source file. If the source is an URL, it downloads it from
        the web.
    dest : path
        Destination path.
    unpack : bool
        If True, it unpacks the data in the dest folder.
    replace_existing : bool
        If True, replaces the existing files.
    """
    try:
        if is_main_process():

            class DownloadProgressBar(tqdm.tqdm):
                """ DownloadProgressBar class."""

                def update_to(self, b=1, bsize=1, tsize=None):
                    """Needed to support multigpu training."""
                    if tsize is not None:
                        self.total = tsize
                    self.update(b * bsize - self.n)

            # Create the destination directory if it doesn't exist
            dest_dir = pathlib.Path(dest).resolve().parent
            dest_dir.mkdir(parents=True, exist_ok=True)
            if "http" not in source:
                shutil.copyfile(source, dest)

            elif not os.path.isfile(dest) or (
                os.path.isfile(dest) and replace_existing
            ):
                print(f"Downloading {source} to {dest}")
                with DownloadProgressBar(
                    unit="B",
                    unit_scale=True,
                    miniters=1,
                    desc=source.split("/")[-1],
                ) as t:
                    urllib.request.urlretrieve(
                        source, filename=dest, reporthook=t.update_to
                    )
            else:
                print(f"{dest} exists. Skipping download")

            # Unpack if necessary
            if unpack:
                if dest_unpack is None:
                    dest_unpack = os.path.dirname(dest)
                print(f"Extracting {dest} to {dest_unpack}")
                shutil.unpack_archive(dest, dest_unpack)
    finally:
        ddp_barrier()


def pad_right_to(
    tensor: torch.Tensor, target_shape: (list, tuple), mode="constant", value=0,
):
    """
    This function takes a torch tensor of arbitrary shape and pads it to target
    shape by appending values on the right.

    Parameters
    ----------
    tensor : input torch tensor
        Input tensor whose dimension we need to pad.
    target_shape : (list, tuple)
        Target shape we want for the target tensor its len must be equal to tensor.ndim
    mode : str
        Pad mode, please refer to torch.nn.functional.pad documentation.
    value : float
        Pad value, please refer to torch.nn.functional.pad documentation.

    Returns
    -------
    tensor : torch.Tensor
        Padded tensor.
    valid_vals : list
        List containing proportion for each dimension of original, non-padded values.
    """
    assert len(target_shape) == tensor.ndim
    pads = []  # this contains the abs length of the padding for each dimension.
    valid_vals = []  # this contains the relative lengths for each dimension.
    i = len(target_shape) - 1  # iterating over target_shape ndims
    j = 0
    while i >= 0:
        assert (
            target_shape[i] >= tensor.shape[i]
        ), "Target shape must be >= original shape for every dim"
        pads.extend([0, target_shape[i] - tensor.shape[i]])
        valid_vals.append(tensor.shape[j] / target_shape[j])
        i -= 1
        j += 1

    tensor = torch.nn.functional.pad(tensor, pads, mode=mode, value=value)

    return tensor, valid_vals


def batch_pad_right(tensors: list, mode="constant", value=0):
    """Given a list of torch tensors it batches them together by padding to the right
    on each dimension in order to get same length for all.

    Parameters
    ----------
    tensors : list
        List of tensor we wish to pad together.
    mode : str
        Padding mode see torch.nn.functional.pad documentation.
    value : float
        Padding value see torch.nn.functional.pad documentation.

    Returns
    -------
    tensor : torch.Tensor
        Padded tensor.
    valid_vals : list
        List containing proportion for each dimension of original, non-padded values.

    """

    if not len(tensors):
        raise IndexError("Tensors list must not be empty")

    if len(tensors) == 1:
        # if there is only one tensor in the batch we simply unsqueeze it.
        return tensors[0].unsqueeze(0), torch.tensor([1.0])

    if not (
        any(
            [tensors[i].ndim == tensors[0].ndim for i in range(1, len(tensors))]
        )
    ):
        raise IndexError("All tensors must have same number of dimensions")

    # FIXME we limit the support here: we allow padding of only the first dimension
    # need to remove this when feat extraction is updated to handle multichannel.
    max_shape = []
    for dim in range(tensors[0].ndim):
        if dim != 0:
            if not all(
                [x.shape[dim] == tensors[0].shape[dim] for x in tensors[1:]]
            ):
                raise EnvironmentError(
                    "Tensors should have same dimensions except for the first one"
                )
        max_shape.append(max([x.shape[dim] for x in tensors]))

    batched = []
    valid = []
    for t in tensors:
        # for each tensor we apply pad_right_to
        padded, valid_percent = pad_right_to(
            t, max_shape, mode=mode, value=value
        )
        batched.append(padded)
        valid.append(valid_percent[0])

    batched = torch.stack(batched)

    return batched, torch.tensor(valid)


def recursive_to(data, *args, **kwargs):
    """Moves data to device, or other type, and handles containers.

    Very similar to torch.utils.data._utils.pin_memory.pin_memory,
    but applies .to() instead.
    """
    if isinstance(data, torch.Tensor):
        return data.to(*args, **kwargs)
    elif isinstance(data, collections.abc.Mapping):
        return {
            k: recursive_to(sample, *args, **kwargs)
            for k, sample in data.items()
        }
    elif isinstance(data, tuple) and hasattr(data, "_fields"):  # namedtuple
        return type(data)(
            *(recursive_to(sample, *args, **kwargs) for sample in data)
        )
    elif isinstance(data, collections.abc.Sequence):
        return [recursive_to(sample, *args, **kwargs) for sample in data]
    elif hasattr(data, "to"):
        return data.to(*args, **kwargs)
    # What should be done with unknown data?
    # For now, just return as they are
    else:
        return data


np_str_obj_array_pattern = re.compile(r"[SaUO]")


def mod_default_collate(batch):
    """Makes a tensor from list of batch values.

    Note that this doesn't need to zip(*) values together
    as PaddedBatch connects them already (by key).

    Here the idea is not to error out.

    This is modified from:
    https://github.com/pytorch/pytorch/blob/c0deb231db76dbea8a9d326401417f7d1ce96ed5/torch/utils/data/_utils/collate.py#L42
    """
    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        out = None
        try:
            if torch.utils.data.get_worker_info() is not None:
                # If we're in a background process, concatenate directly into a
                # shared memory tensor to avoid an extra copy
                numel = sum([x.numel() for x in batch])
                storage = elem.storage()._new_shared(numel)
                out = elem.new(storage)
            return torch.stack(batch, 0, out=out)
        except RuntimeError:  # Unequal size:
            return batch
    elif (
        elem_type.__module__ == "numpy"
        and elem_type.__name__ != "str_"
        and elem_type.__name__ != "string_"
    ):
        try:
            if (
                elem_type.__name__ == "ndarray"
                or elem_type.__name__ == "memmap"
            ):
                # array of string classes and object
                if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                    return batch
                return mod_default_collate([torch.as_tensor(b) for b in batch])
            elif elem.shape == ():  # scalars
                return torch.as_tensor(batch)
        except RuntimeError:  # Unequal size
            return batch
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(elem, int):
        return torch.tensor(batch)
    else:
        return batch



# Copyright (c) 2022, Intel Corporation

## refer to https://github.com/megvii-research/mdistiller

# MIT License

# Copyright (c) 2022 MEGVII Research

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



# detectron2

# Copyright 2020 - present, Facebook, Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# RepDistiller

# Copyright (c) 2020, Yonglong Tian

import torch
import torch.nn as nn
import torch.nn.functional as F
from .basic_distiller import BasicDistiller

def dkd_loss(logits_student, logits_teacher, target, alpha, beta, temperature):
    gt_mask = _get_gt_mask(logits_student, target)
    other_mask = _get_other_mask(logits_student, target)
    pred_student = F.softmax(logits_student / temperature, dim=1)
    pred_teacher = F.softmax(logits_teacher / temperature, dim=1)
    pred_student = cat_mask(pred_student, gt_mask, other_mask)
    pred_teacher = cat_mask(pred_teacher, gt_mask, other_mask)
    log_pred_student = torch.log(pred_student)
    tckd_loss = (
        F.kl_div(log_pred_student, pred_teacher, size_average=False)
        * (temperature**2)
        / target.shape[0]
    )
    pred_teacher_part2 = F.softmax(
        logits_teacher / temperature - 1000.0 * gt_mask, dim=1
    )
    log_pred_student_part2 = F.log_softmax(
        logits_student / temperature - 1000.0 * gt_mask, dim=1
    )
    nckd_loss = (
        F.kl_div(log_pred_student_part2, pred_teacher_part2, size_average=False)
        * (temperature**2)
        / target.shape[0]
    )
    return alpha * tckd_loss + beta * nckd_loss


def _get_gt_mask(logits, target):
    target = target.reshape(-1)
    mask = torch.zeros_like(logits).scatter_(1, target.unsqueeze(1), 1).bool()
    return mask


def _get_other_mask(logits, target):
    target = target.reshape(-1)
    mask = torch.ones_like(logits).scatter_(1, target.unsqueeze(1), 0).bool()
    return mask


def cat_mask(t, mask1, mask2):
    t1 = (t * mask1).sum(dim=1, keepdims=True)
    t2 = (t * mask2).sum(1, keepdims=True)
    rt = torch.cat([t1, t2], dim=1)
    return rt


class DKD(BasicDistiller):
    """Decoupled Knowledge Distillation(CVPR 2022)"""
    def __init__(self, pretrained_model, alpha, beta, temperature, warmup, is_frozen=True, use_saved_logits=True, topk=0, num_classes=10, teacher_type=None):
        ''' Init method.

        :param pretrained_model: the pretrained model as teacher
        :param alpha: the alpha for DKD 
        :param beta: the beta for DKD 
        :param temperature: the temperature for DKD 
        :param warmup: warmup epoches for DKD
        :param is_frozen: whether frozen teacher when training
        :param use_saved_logits: whether train with pre-saved logits
        :param topk: if use logits, save top k logits, 0 means save all logits
        :param num_classes: num of classification classes
        :param teacher_type: teacher model type
        '''
        super(DKD, self).__init__(pretrained_model, is_frozen, use_saved_logits, topk, num_classes, teacher_type)
        self.alpha = alpha
        self.beta = beta
        self.temperature = temperature
        self.warmup = warmup

    def loss(self,teacher_logits, student_logits,**kwargs):
        ''' Loss function.

        :param teacher_logits: the teacher logits
        :param student_logits: the student logits
        '''
        distiller_loss = min(kwargs["epoch"] / self.warmup, 1.0) * dkd_loss(
                            student_logits, teacher_logits,kwargs["target"],
                            self.alpha, self.beta, self.temperature,)
        return distiller_loss



import torch
import logging


logger = logging.getLogger(__name__)

def load_torch_model(obj, path, device):
    incompatible_keys = obj.load_state_dict(
        torch.load(path, map_location=device), strict=False
    )
    for missing_key in incompatible_keys.missing_keys:
        logger.warning(
            f"During parameter transfer to {obj} loading from "
            + f"{path}, the transferred parameters did not have "
            + f"parameters for the key: {missing_key}"
        )
    for unexpected_key in incompatible_keys.unexpected_keys:
        logger.warning(
            f"During parameter transfer to {obj} loading from "
            + f"{path}, the object could not use the parameters loaded "
            + f"with the key: {unexpected_key}"
        )

def load_spm(obj, path):
    obj.load(str(path))


import torch
from torch import nn
from typing import Optional

import os, sys

from e2eAIOK.DeNas.module.asr.linear import Linear
from e2eAIOK.DeNas.asr.TransformerBase import (
    get_lookahead_mask,
    get_key_padding_mask,
    NormalizedEmbedding,
    PositionalEncoding
)
from e2eAIOK.DeNas.module.asr.encoder import TransformerEncoder
from e2eAIOK.DeNas.module.asr.decoder import TransformerDecoder
from e2eAIOK.DeNas.asr.data.dataio.dataio import length_to_mask


class TransformerASRSuper(nn.Module):
    def __init__(
        self,
        tgt_vocab,
        input_size,
        d_model=512,
        encoder_heads=[4],
        nhead=8,
        num_encoder_layers=6,
        num_decoder_layers=6,
        mlp_ratio=[4],
        d_ffn=2048,
        dropout=0.1,
        activation=nn.ReLU,
        normalize_before=False,
        max_length: Optional[int] = 2500,
    ):
        super().__init__()
        self.encoder_kdim = None
        self.encoder_vdim = None
        self.decoder_kdim = None
        self.decoder_vdim = None

        self.positional_encoding = PositionalEncoding(d_model, max_length)

        # initialize the encoder
        self.encoder = TransformerEncoder(
            nhead=encoder_heads,
            num_layers=num_encoder_layers,
            mlp_ratio=mlp_ratio,
            d_model=d_model,
            dropout=dropout,
            activation=activation,
            normalize_before=normalize_before,
            kdim=self.encoder_kdim,
            vdim=self.encoder_vdim,
        )
        # initialize the decoder
        self.decoder = TransformerDecoder(
            num_layers=num_decoder_layers,
            nhead=nhead,
            d_ffn=d_ffn,
            d_model=d_model,
            dropout=dropout,
            activation=activation,
            normalize_before=normalize_before,
            kdim=self.decoder_kdim,
            vdim=self.decoder_vdim,
        )

        self.custom_src_module = nn.ModuleList([
            Linear(
                input_size=input_size,
                n_neurons=d_model,
                bias=True,
            ),
            torch.nn.Dropout(dropout),
        ])
        self.custom_tgt_module = nn.ModuleList([
            NormalizedEmbedding(d_model, tgt_vocab)
        ])

        # reset parameters using xavier_normal_
        self._init_params()

    def forward(self, src, tgt, wav_len=None, pad_idx=0):
        encoder_out, src_key_padding_mask = self.encode(src, wav_len)

        tgt_key_padding_mask = get_key_padding_mask(tgt, pad_idx=pad_idx)

        src_mask = None
        tgt_mask = get_lookahead_mask(tgt)

        for layer in self.custom_tgt_module:
            tgt = layer(tgt)

        tgt = tgt + self.positional_encoding(tgt)
        pos_embs_target = None
        pos_embs_encoder = None

        decoder_out, _, _ = self.decoder(
            tgt=tgt,
            memory=encoder_out,
            memory_mask=src_mask,
            tgt_mask=tgt_mask,
            tgt_key_padding_mask=tgt_key_padding_mask,
            memory_key_padding_mask=src_key_padding_mask,
            pos_embs_tgt=pos_embs_target,
            pos_embs_src=pos_embs_encoder,
        )

        return encoder_out, decoder_out

    def make_masks(self, src, tgt, wav_len=None, pad_idx=0):
        src_key_padding_mask = None
        if wav_len is not None:
            abs_len = torch.round(wav_len * src.shape[1])
            src_key_padding_mask = ~length_to_mask(abs_len).bool()

        tgt_key_padding_mask = get_key_padding_mask(tgt, pad_idx=pad_idx)

        src_mask = None
        tgt_mask = get_lookahead_mask(tgt)
        return src_key_padding_mask, tgt_key_padding_mask, src_mask, tgt_mask

    @torch.no_grad()
    def decode(self, tgt, encoder_out, enc_len=None):
        tgt_mask = get_lookahead_mask(tgt)
        src_key_padding_mask = None
        if enc_len is not None:
            src_key_padding_mask = (1 - length_to_mask(enc_len)).bool()

        for layer in self.custom_tgt_module:
            tgt = layer(tgt)
        tgt = tgt + self.positional_encoding(tgt)  # add the encodings here
        pos_embs_target = None
        pos_embs_encoder = None

        prediction, self_attns, multihead_attns = self.decoder(
            tgt,
            encoder_out,
            tgt_mask=tgt_mask,
            memory_key_padding_mask=src_key_padding_mask,
            pos_embs_tgt=pos_embs_target,
            pos_embs_src=pos_embs_encoder,
        )
        return prediction, multihead_attns[-1]

    def encode(self, src, wav_len=None):
        # reshape the src vector to [Batch, Time, Fea] if a 4d vector is given
        if src.dim() == 4:
            bz, t, ch1, ch2 = src.shape
            src = src.reshape(bz, t, ch1 * ch2)

        src_key_padding_mask = None
        if wav_len is not None:
            abs_len = torch.round(wav_len * src.shape[1])
            src_key_padding_mask = ~length_to_mask(abs_len).bool()

        for layer in self.custom_src_module:
            src = layer(src)

        src = src + self.positional_encoding(src)
        pos_embs_source = None

        encoder_out, _ = self.encoder(
            src=src,
            src_key_padding_mask=src_key_padding_mask,
            pos_embs=pos_embs_source,
        )
        return encoder_out, src_key_padding_mask

    def _init_params(self):
        for p in self.parameters():
            if p.dim() > 1:
                torch.nn.init.xavier_normal_(p)

    def calc_sampled_param_num(self):
        numel = 0
        numel += self.encoder.calc_sampled_param_num()
        numel += self.custom_src_module[0].calc_sampled_param_num()
        return numel

    @classmethod
    def gen_transformer(config):
        model = TransformerASRSuper(
            input_size = config["input_size"],
            tgt_vocab = config["output_neurons"],
            d_model = config["d_model"],
            encoder_heads = config["encoder_heads"],
            nhead = config["nhead"],
            num_encoder_layers = config["num_encoder_layers"],
            num_decoder_layers = config["num_decoder_layers"],
            mlp_ratio = config["mlp_ratio"],
            d_ffn = config["d_ffn"],
            dropout = config["transformer_dropout"],
            activation = torch.nn.GELU,
            normalize_before = True
        )
        return model